#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


from __future__ import annotations

import pathlib
import time
from contextlib import contextmanager
from logging import getLogger
from typing import TYPE_CHECKING, Optional

from PySide6.QtCore import (Qt, QSize, )
from PySide6.QtGui import (QAction, QIcon, QImage, QKeySequence)
from PySide6.QtPrintSupport import QPrintDialog, QPrinter, QPrintPreviewDialog
from PySide6.QtWidgets import (QApplication, QDialog, QFileDialog, QGraphicsScene,
                               QLabel, QMainWindow, QDialogButtonBox,
                               QMenu, QToolBar, QWidget)

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.libBackEnd as libb
import revedaEditor.backend.libraryModelView as lmview
import revedaEditor.backend.processManager as prm
import revedaEditor.gui.alignItems as alg
import revedaEditor.gui.helpBrowser as hlp
import revedaEditor.gui.propertyDialogues as pdlg
import revedaEditor.resources.resources  # noqa: F401
from revedaEditor.backend.startThread import startThread
from revedaEditor.gui.editorViews import editorView
from revedaEditor.scenes.editorScene import editorScene

if TYPE_CHECKING:
    from revedaEditor.gui.editorTypes import EditorContainer


class editorWindow(QMainWindow):
    """
    Base class for editor windows.
    """
    MAIN_LOGGER = "reveda"
    MAJOR_GRID_DEFAULT = 20
    SNAP_GRID_DEFAULT = 10

    def __init__(self, viewItem: libb.viewItem, libraryDict: dict,
                 libraryView: lmview.BaseDesignLibrariesView) -> None:
        super().__init__()

        self.viewItem = viewItem
        self.file: pathlib.Path = self.viewItem.data(
            Qt.ItemDataRole.UserRole + 2)  # pathlib Path object
        self.cellItem: libb.cellItem = self.viewItem.parent()
        self.cellName = self.cellItem.cellName
        self.libItem: libb.libraryItem = self.cellItem.parent()
        self.libName: str = self.libItem.libraryName
        self.viewName: str = self.viewItem.viewName
        self.libraryDict = libraryDict
        self.libraryView = libraryView
        self.parentEditor: Optional[editorWindow] = None
        self.parentObj = None  # type symbol or layoutInstance
        self._app = QApplication.instance()  # main application pointer
        self.appMainW = self._app.appMainW  # main window pointer
        self.logger = getLogger(self.MAIN_LOGGER)
        self.switchViewList = self.appMainW.switchViewList
        self.stopViewList = self.appMainW.stopViewList
        self.statusLine = self.statusBar()
        self.messageLine = QLabel()  # message line
        self.statusLine.insertPermanentWidget(0, self.messageLine)
        self.majorGrid = self.MAJOR_GRID_DEFAULT  # dot/line grid spacing
        self.snapGrid = self.SNAP_GRID_DEFAULT  # snapping grid size
        self.snapTuple = (self.snapGrid, self.snapGrid)
        self.processManager = prm.ProcessManager(3, self)
        self.init_UI()
        self._createSignalConnections()

    def __repr__(self):
        return f'editorWindow({self.libName}-{self.cellName}-{self.viewName})'

    def init_UI(self) -> None:
        self.centralW: EditorContainer = editorContainer(self)
        self.setCentralWidget(self.centralW)
        self.resize(1600, 800)
        self._createActions()
        self._createMenuBar()
        self._createToolBars()
        self._addActions()
        self._createTriggers()
        self._createShortcuts()

    def _createMenuBar(self):
        """
        Creates the menu bar for the editor.

        """
        self.editorMenuBar = self.menuBar()
        self.editorMenuBar.setNativeMenuBar(False)
        # Returns QMenu object.
        self.menuFile = self.editorMenuBar.addMenu("&File")
        self.menuView = self.editorMenuBar.addMenu("&View")
        self.menuEdit = self.editorMenuBar.addMenu("&Edit")
        self.menuCreate = self.editorMenuBar.addMenu("C&reate")
        self.menuOptions = self.editorMenuBar.addMenu("&Options")
        self.menuCheck = self.editorMenuBar.addMenu("&Check")
        self.menuTools = self.editorMenuBar.addMenu("&Tools")
        # self.menuWindow = self.editorMenuBar.addMenu("&Window")
        self.menuUtilities = self.editorMenuBar.addMenu("&Utilities")
        self.menuHelp = self.editorMenuBar.addMenu("&Help")

    def _createActions(self):
        checkCellIcon = QIcon(":/icons/document-task.png")
        self.checkCellAction = QAction(checkCellIcon, "Check-Save", self)

        saveCellIcon = QIcon(":/icons/document--plus.png")
        self.saveCellAction = QAction(saveCellIcon, "Save", self)

        self.readOnlyCellIcon = QIcon(":/icons/lock.png")
        self.readOnlyCellAction = QAction("Read Only", self)
        self.readOnlyCellAction.setCheckable(True)

        updateCellIcon = QIcon(":/icons/document-xaml.png")
        self.updateCellAction = QAction(updateCellIcon, "Update Design", self)
        self.updateCellAction.setToolTip("Reread all the cells in the design")

        printIcon = QIcon(":/icons/printer--arrow.png")
        self.printAction = QAction(printIcon, "Print...", self)
        self.printAction.setToolTip("Print the current design")

        printPreviewIcon = QIcon(":/icons/printer--arrow.png")
        self.printPreviewAction = QAction(printPreviewIcon, "Print Preview...",
                                          self)
        self.printPreviewAction.setToolTip("Preview the current design output")

        exportImageIcon = QIcon(":/icons/image-export.png")
        self.exportImageAction = QAction(exportImageIcon, "Export...", self)
        self.exportImageAction.setToolTip("Export the current design as an image")

        exitIcon = QIcon(":/icons/external.png")
        self.exitAction = QAction(exitIcon, "Close Window", self)
        self.exitAction.setShortcut("Ctrl+Q")
        self.exitAction.setToolTip("Close the current window")

        fitIcon = QIcon(":/icons/zone.png")
        self.fitAction = QAction(fitIcon, "Fit to Window", self)
        self.fitAction.setToolTip("Fit the design to the window")

        zoomInIcon = QIcon(":/icons/zone-resize.png")
        self.zoomInAction = QAction(zoomInIcon, "Zoom In", self)
        self.zoomInAction.setToolTip("Zoom in on the design")

        zoomOutIcon = QIcon(":/icons/zone-resize-actual.png")
        self.zoomOutAction = QAction(zoomOutIcon, "Zoom Out", self)
        self.zoomOutAction.setToolTip("Zoom out on the design")

        panIcon = QIcon(":/icons/zone--arrow.png")
        self.panAction = QAction(panIcon, "Pan View", self)
        self.panAction.setToolTip("Pan the design")

        redrawIcon = QIcon(":/icons/arrow-circle.png")
        self.redrawAction = QAction(redrawIcon, "Redraw", self)
        self.redrawAction.setToolTip("Redraw the design on the screen")

        rulerIcon = QIcon(":/icons/ruler.png")
        self.rulerAction = QAction(rulerIcon, "Add Ruler", self)
        self.rulerAction.setToolTip("Add a ruler to the layout")

        delRulerIcon = QIcon(":/icons/ruler--minus.png")
        self.delRulerAction = QAction(delRulerIcon, "Delete Rulers", self)
        self.delRulerAction.setToolTip("Delete all the rulers from the layout")

        alignIcon = QIcon(":/icons/layers-alignment-middle.png")
        self.alignItemsAction = QAction(alignIcon, "Align Items...", self)
        self.alignItemsAction.setToolTip("Align selected items")

        # display options
        dispConfigIcon = QIcon(":/icons/grid-snap-dot.png")
        self.dispConfigAction = QAction(dispConfigIcon, "Display Config...", self)
        self.dispConfigAction.setToolTip("Configure the display options")

        selectConfigIcon = QIcon(":/icons/zone-select.png")
        self.selectConfigAction = QAction(selectConfigIcon, "Selection Config...",
                                          self)
        self.selectConfigAction.setToolTip("Configure the selection options")

        panZoomConfigIcon = QIcon(":/icons/selection-resize.png")
        self.panZoomConfigAction = QAction(panZoomConfigIcon,
                                           "Pan/Zoom Config...", self)
        self.panZoomConfigAction.setToolTip("Configure the pan/zoom options")

        undoIcon = QIcon(":/icons/arrow-circle-315-left.png")
        self.undoAction = QAction(undoIcon, "Undo", self)
        self.undoAction.setToolTip("Undo the last action")

        redoIcon = QIcon(":/icons/arrow-circle-225.png")
        self.redoAction = QAction(redoIcon, "Redo", self)
        self.redoAction.setToolTip("Redo the last undone action")

        yankIcon = QIcon(":/icons/node-insert.png")
        self.yankAction = QAction(yankIcon, "Yank", self)

        pasteIcon = QIcon(":/icons/clipboard-paste.png")
        self.pasteAction = QAction(pasteIcon, "Paste", self)
        self.pasteAction.setToolTip("Paste the contents of the clipboard")

        deleteIcon = QIcon(":/icons/node-delete.png")
        self.deleteAction = QAction(deleteIcon, "Delete", self)
        self.deleteAction.setToolTip("Delete selected items")

        copyIcon = QIcon(":/icons/document-copy.png")
        self.copyAction = QAction(copyIcon, "Copy", self)
        self.copyAction.setToolTip("Copy selected items")

        moveIcon = QIcon(":/icons/arrow-move.png")
        self.moveAction = QAction(moveIcon, "Move", self)
        self.moveAction.setToolTip("Move selected items")

        moveByIcon = QIcon(":/icons/arrow-transition.png")
        self.moveByAction = QAction(moveByIcon, "Move By ...", self)
        self.moveAction.setToolTip("Move selected items by an offset")

        moveOriginIcon = QIcon(":/icons/arrow-skip.png")
        self.moveOriginAction = QAction(moveOriginIcon, "Move Origin", self)
        self.moveOriginAction.setToolTip("Move the origin of the design")

        stretchIcon = QIcon(":/icons/fill.png")
        self.stretchAction = QAction(stretchIcon, "Stretch", self)
        self.stretchAction.setToolTip("Stretch item")

        rotateIcon = QIcon(":/icons/arrow-circle.png")
        self.rotateAction = QAction(rotateIcon, "Rotate...", self)
        self.rotateAction.setToolTip("Rotate item")

        scaleIcon = QIcon(":/icons/selection-resize.png")
        self.scaleAction = QAction(scaleIcon, "Scale...", self)
        self.scaleAction.setToolTip("Scale item")

        verticalFlipIcon = QIcon(":/icons/layer-flip-vertical.png")
        self.verticalFlipAction = QAction(verticalFlipIcon, "Vertical Flip", self)
        self.verticalFlipAction.setToolTip("Vertical Flip")

        horizontalFlipIcon = QIcon(":/icons/layer-flip.png")
        self.horizontalFlipAction = QAction(horizontalFlipIcon, "Horizontal Flip",
                                            self)
        self.horizontalFlipAction.setToolTip("Horizontal Flip")

        netNameIcon = QIcon(":/icons/node-design.png")
        self.netNameAction = QAction(netNameIcon, "Net Name...", self)
        self.netNameAction.setToolTip("Set net name")

        # create label action but do not add to any menu.
        createLabelIcon = QIcon(":/icons/tag-label-yellow.png")
        self.createLabelAction = QAction(createLabelIcon, "Create Label...", self)
        self.createLabelAction.setToolTip("Create Label")

        createPinIcon = QIcon(":/icons/pin.png")
        self.createPinAction = QAction(createPinIcon, "Create Pin...", self)
        self.createPinAction.setToolTip("Create Pin")

        goUpIcon = QIcon(":/icons/arrow-step-out.png")
        self.goUpAction = QAction(goUpIcon, "Go Up", self)
        self.goUpAction.setToolTip("Go up a level in design hierarchy")

        goDownIcon = QIcon(":/icons/arrow-step.png")
        self.goDownAction = QAction(goDownIcon, "Go Down", self)
        self.goDownAction.setToolTip("Go down a level in design hierarchy")

        self.selectAllIcon = QIcon(":/icons/node-select-all.png")
        self.selectAllAction = QAction(self.selectAllIcon, "Select All", self)
        self.selectAllAction.setToolTip("Select all items in the design")

        deselectAllIcon = QIcon(":/icons/node.png")
        self.deselectAllAction = QAction(deselectAllIcon, "Unselect All", self)
        self.deselectAllAction.setToolTip("Unselect all items in the design")

        objPropIcon = QIcon(":/icons/property-blue.png")
        self.objPropAction = QAction(objPropIcon, "Object Properties...", self)
        self.objPropAction.setToolTip("Configure object properties")

        viewPropIcon = QIcon(":/icons/property.png")
        self.viewPropAction = QAction(viewPropIcon, "Cellview Properties...",
                                      self)
        self.viewPropAction.setToolTip("Configure Cellview Properties")

        viewCheckIcon = QIcon(":/icons/ui-check-box.png")
        self.viewCheckAction = QAction(viewCheckIcon, "Check CellView", self)
        self.viewCheckAction.setToolTip("Check Cellview")

        viewErrorsIcon = QIcon(":/icons/report--exclamation.png")
        self.viewErrorsAction = QAction(viewErrorsIcon, "View Errors...", self)
        self.viewErrorsAction.setToolTip("View Errros")

        deleteErrorsIcon = QIcon(":/icons/report--minus.png")
        self.deleteErrorsAction = QAction(deleteErrorsIcon, "Delete Errors...",
                                          self)
        self.deleteErrorsAction.setToolTip("Delete Errros")

        netlistIcon = QIcon(":/icons/script-text.png")
        self.netlistAction = QAction(netlistIcon, "Create Netlist...", self)
        self.netlistAction.setToolTip("Create Netlist")

        createLineIcon = QIcon(":/icons/layer-shape-line.png")
        self.createLineAction = QAction(createLineIcon, "Create Line...", self)
        self.createLineAction.setToolTip("Create Line")

        createRectIcon = QIcon(":/icons/layer-shape.png")
        self.createRectAction = QAction(createRectIcon, "Create Rectangle...",
                                        self)
        self.createRectAction.setToolTip("Create Rectangle")

        createPolyIcon = QIcon(":/icons/layer-shape-polygon.png")
        self.createPolygonAction = QAction(createPolyIcon, "Create Polygon...",
                                           self)
        self.createPolygonAction.setToolTip("Create Polygon")

        createCircleIcon = QIcon(":/icons/layer-shape-ellipse.png")
        self.createCircleAction = QAction(createCircleIcon, "Create Circle...",
                                          self)
        self.createCircleAction.setToolTip("Create Circle")

        createArcIcon = QIcon(":/icons/layer-shape-polyline.png")
        self.createArcAction = QAction(createArcIcon, "Create Arc...", self)
        self.createArcAction.setToolTip("Create Arc")

        createViaIcon = QIcon(":/icons/layer-mask.png")
        self.createViaAction = QAction(createViaIcon, "Create Via...", self)
        self.createViaAction.setToolTip("Create Via")

        createInstIcon = QIcon(":/icons/block--plus.png")
        self.createInstAction = QAction(createInstIcon, "Create Instance...",
                                        self)
        self.createInstAction.setToolTip("Create Instance")

        self.createNetAction = QAction(createLineIcon, "Create Net...", self)
        self.createNetAction.setToolTip("Create Net")

        self.createPathAction = QAction(createLineIcon, "Create Path...", self)
        self.createPathAction.setToolTip("Create Path")

        createBusIcon = QIcon(":/icons/node-select-all.png")
        self.createBusAction = QAction(createBusIcon, "Create Bus...", self)
        self.createBusAction.setToolTip("Create Bus")

        createSymbolIcon = QIcon(":/icons/application-block.png")
        self.createSymbolAction = QAction(createSymbolIcon, "Create Symbol...",
                                          self)
        self.createSymbolAction.setToolTip("Create Symbol from Cellview")

        createTextIcon = QIcon(":icons/sticky-note-text.png")
        self.createTextAction = QAction(createTextIcon, "Create Text...", self)
        self.createTextAction.setToolTip("Create Text")

        # selection Actions
        selectDeviceIcon = QIcon(":icons/target.png")
        self.selectDeviceAction = QAction(selectDeviceIcon, "Select Devices",
                                          self)
        self.selectDeviceAction.setToolTip("Select Devices Only")

        selectNetIcon = QIcon(":icons/pencil--plus.png")
        self.selectNetAction = QAction(selectNetIcon, "Select Nets", self)
        self.selectNetAction.setToolTip("Select Nets Only")

        self.selectWireAction = QAction(selectNetIcon, "Select Wires", self)
        self.selectWireAction.setToolTip("Select Wires Only")

        selectPinIcon = QIcon(":/icons/pin--plus.png")
        self.selectPinAction = QAction(selectPinIcon, "Select Pins", self)
        self.selectPinAction.setToolTip("Select Pins Only")

        removeSelectFilterIcon = QIcon(":icons/eraser.png")
        self.removeSelectFilterAction = QAction(removeSelectFilterIcon,
                                                "Remove Select Filters", self)
        self.removeSelectFilterAction.setToolTip("Remove Selection Filters")

        ignoreIcon = QIcon(":/icons/minus-circle.png")
        self.ignoreAction = QAction(ignoreIcon, "Ignore", self)
        self.ignoreAction.setToolTip("Ignore selected cell")

        helpIcon = QIcon(":/icons/document-arrow.png")
        self.helpAction = QAction(helpIcon, "Help...", self)
        self.helpAction.setToolTip("Help")

        self.aboutIcon = QIcon(":/icons/information.png")
        self.aboutAction = QAction(self.aboutIcon, "About", self)

    def _createToolBars(self):
        # Create tools bar called "main toolbar"
        self.toolbar = QToolBar("Main Toolbar", self)
        self.toolbar.setIconSize(QSize(32, 32))
        # place toolbar at top
        self.addToolBar(self.toolbar)
        self.toolbar.addAction(self.saveCellAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.printAction)
        self.toolbar.addAction(self.exportImageAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.undoAction)
        self.toolbar.addAction(self.redoAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.deleteAction)
        self.toolbar.addAction(self.moveAction)
        self.toolbar.addAction(self.copyAction)
        self.toolbar.addAction(self.stretchAction)
        self.toolbar.addAction(self.rotateAction)
        self.toolbar.addAction(self.horizontalFlipAction)
        self.toolbar.addAction(self.verticalFlipAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.fitAction)
        self.toolbar.addAction(self.zoomInAction)
        self.toolbar.addAction(self.zoomOutAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.objPropAction)

    def _addActions(self):
        # file menu
        self.menuFile.addAction(self.checkCellAction)
        self.menuFile.addAction(self.saveCellAction)
        self.menuFile.addAction(self.updateCellAction)
        self.menuFile.addAction(self.printAction)
        self.menuFile.addAction(self.printPreviewAction)
        self.menuFile.addAction(self.exportImageAction)
        self.menuFile.addAction(self.exitAction)
        # view menu
        self.menuView.addAction(self.fitAction)
        self.menuView.addAction(self.zoomInAction)
        self.menuView.addAction(self.zoomOutAction)
        self.menuView.addAction(self.panAction)
        self.menuView.addAction(self.redrawAction)
        # self.menuView.addAction(self.panZoomConfigAction)
        # edit menu
        self.menuEdit.addAction(self.undoAction)
        self.menuEdit.addAction(self.redoAction)
        # self.menuEdit.addAction(self.yankAction)
        self.menuEdit.addAction(self.pasteAction)
        self.menuEdit.addAction(self.deleteAction)
        self.menuEdit.addAction(self.copyAction)
        self.menuEdit.addAction(self.moveAction)
        self.menuEdit.addAction(self.moveByAction)
        self.menuEdit.addAction(self.moveOriginAction)
        self.menuEdit.addAction(self.stretchAction)
        self.menuEdit.addAction(self.rotateAction)
        self.menuEdit.addAction(self.horizontalFlipAction)
        self.menuEdit.addAction(self.verticalFlipAction)
        self.menuEdit.addAction(self.alignItemsAction)
        self.selectMenu = QMenu('Selection', self)
        self.selectMenu.setIcon(QIcon('icons/node-select.png'))
        self.menuEdit.addMenu(self.selectMenu)
        self.selectMenu.addAction(self.selectAllAction)
        self.selectMenu.addAction(self.deselectAllAction)
        self.menuTools.addAction(self.readOnlyCellAction)
        # self.menuCheck.addAction(self.viewCheckAction)
        self.menuOptions.addAction(self.dispConfigAction)
        self.menuOptions.addAction(self.selectConfigAction)
        self.menuHelp.addAction(self.helpAction)
        self.menuHelp.addAction(self.aboutAction)

    def helpClick(self):
        helpBrowser = hlp.helpBrowser(self)
        helpBrowser.show()

    def aboutClick(self):
        abtDlg = hlp.aboutDialog(self)
        abtDlg.show()

    def _createTriggers(self):
        self.checkCellAction.triggered.connect(self.checkSaveCell)
        self.saveCellAction.triggered.connect(self.saveCell)
        self.readOnlyCellAction.triggered.connect(self.readOnlyCellClick)
        self.updateCellAction.triggered.connect(self.updateDesignScene)
        self.printAction.triggered.connect(self.printClick)
        self.printPreviewAction.triggered.connect(self.printPreviewClick)
        self.exportImageAction.triggered.connect(self.imageExportClick)
        self.exitAction.triggered.connect(self.closeWindow)
        self.fitAction.triggered.connect(self.fitToWindow)
        self.redrawAction.triggered.connect(self.redraw)
        self.zoomInAction.triggered.connect(self.zoomIn)
        self.zoomOutAction.triggered.connect(self.zoomOut)
        self.panAction.triggered.connect(self.panView)
        self.dispConfigAction.triggered.connect(self.dispConfigEdit)
        self.selectConfigAction.triggered.connect(self.selectConfigEdit)
        self.stretchAction.triggered.connect(self.stretchClick)
        self.moveOriginAction.triggered.connect(self.moveOrigin)
        self.selectAllAction.triggered.connect(self.selectAllClick)
        self.deselectAllAction.triggered.connect(self.deselectAllClick)
        self.deleteAction.triggered.connect(self.deleteClick)
        self.copyAction.triggered.connect(self.copyClick)
        self.undoAction.triggered.connect(self.undoClick)
        self.redoAction.triggered.connect(self.redoClick)
        self.moveAction.triggered.connect(self.moveClick)
        self.moveByAction.triggered.connect(self.moveByClick)
        self.rotateAction.triggered.connect(self.rotateItemClick)
        self.verticalFlipAction.triggered.connect(self.verticalFlipClick)
        self.horizontalFlipAction.triggered.connect(self.horizontalFlipClick)
        self.alignItemsAction.triggered.connect(self.alignItemsClick)
        self.goUpAction.triggered.connect(self.goUpHierarchy)
        self.helpAction.triggered.connect(self.helpClick)
        self.aboutAction.triggered.connect(self.aboutClick)

    def _createShortcuts(self):
        self.redoAction.setShortcut("Shift+U")
        self.undoAction.setShortcut(Qt.Key.Key_U)
        self.objPropAction.setShortcut(Qt.Key.Key_Q)
        self.copyAction.setShortcut(Qt.Key.Key_C)
        self.rotateAction.setShortcut("Ctrl+R")
        self.createTextAction.setShortcut("Shift+L")
        self.fitAction.setShortcut(Qt.Key.Key_F)
        self.deleteAction.setShortcut(QKeySequence.StandardKey.Delete)
        self.selectAllAction.setShortcut("Ctrl+A")
        self.stretchAction.setShortcut(Qt.Key.Key_S)
        self.alignItemsAction.setShortcut("Shift+A")

    def _editorContextMenu(self):
        self.centralW.scene.itemContextMenu.addAction(self.copyAction)
        self.centralW.scene.itemContextMenu.addAction(self.moveAction)
        self.centralW.scene.itemContextMenu.addAction(self.moveByAction)
        self.centralW.scene.itemContextMenu.addAction(self.verticalFlipAction)
        self.centralW.scene.itemContextMenu.addAction(self.horizontalFlipAction)
        self.centralW.scene.itemContextMenu.addAction(self.rotateAction)
        self.centralW.scene.itemContextMenu.addAction(self.deleteAction)
        self.centralW.scene.itemContextMenu.addAction(self.objPropAction)
        self.centralW.scene.itemContextMenu.addAction(self.selectAllAction)
        self.centralW.scene.itemContextMenu.addAction(self.deselectAllAction)

    def checkSaveCell(self):
        pass

    def saveCell(self):
        pass

    def dispConfigEdit(self):

        dcd = pdlg.displayConfigDialog(self)
        dcd.majorGridEntry.setText(str(self.majorGrid))
        dcd.snapGridEdit.setText(str(self.snapGrid))
        if dcd.exec() == QDialog.DialogCode.Accepted:
            self.configureGridSettings(
                (int(dcd.majorGridEntry.text()), int(dcd.snapGridEdit.text())))
            if dcd.dotType.isChecked():
                self.centralW.view.gridbackg = True
                self.centralW.view.linebackg = False
            elif dcd.lineType.isChecked():
                self.centralW.view.gridbackg = False
                self.centralW.view.linebackg = True
            else:
                self.centralW.view.gridbackg = False
                self.centralW.view.linebackg = False

    def configureGridSettings(self, gridSettings: tuple[int, int]) -> None:
        """Configure grid settings from decoded data."""
        try:

            # Update editor window
            self.majorGrid, self.snapGrid = gridSettings
            # majorG, snapG = gridSettings
            # print(majorG, snapG)
            self.snapTuple = (self.snapGrid, self.snapGrid)

            # Update scene and view if they exist and have these attributes
            if hasattr(self, 'centralW') and self.centralW:
                for obj in (self.centralW.scene, self.centralW.view):
                    if hasattr(obj, 'majorGrid'):
                        obj.majorGrid = self.majorGrid
                    if hasattr(obj, 'snapGrid'):
                        obj.snapGrid = self.snapGrid
                    if hasattr(obj, 'snapTuple'):
                        obj.snapTuple = (self.snapGrid, self.snapGrid)
                self.centralW.scene.invalidate(self.centralW.scene.sceneRect(),
                                               QGraphicsScene.SceneLayer.BackgroundLayer)
        except Exception as e:
            self.logger.info(f'Problem with grid settings: {e}')

    def selectConfigEdit(self):
        scd = pdlg.selectConfigDialogue(self)
        if self.centralW.scene.partialSelection:
            scd.partialSelection.setChecked(True)
        else:
            scd.fullSelection.setChecked(True)
        if scd.exec() == QDialog.DialogCode.Accepted:
            self.centralW.scene.partialSelection = scd.partialSelection.isChecked()

    def readOnlyCellClick(self):
        self.centralW.scene.readOnly = self.readOnlyCellAction.isChecked()

    def updateDesignScene(self):
        # reloadScene() is implemented editor scenes
        self.messageLine.setText("Reloading design.")
        self.centralW.scene.reloadScene()

    def printClick(self):
        dlg = QPrintDialog(self)
        if dlg.exec() == QDialog.DialogCode.Accepted:
            printer = dlg.printer()
            printRunner = startThread(self.centralW.view.printView(printer))
            self.appMainW.threadPool.start(printRunner)
            self.logger.info(
                "Printing started")  # self.centralW.view.printView(printer)

    def printPreviewClick(self):
        printer = QPrinter(QPrinter.PrinterMode.ScreenResolution)
        printer.setOutputFormat(QPrinter.OutputFormat.PdfFormat)
        ppdlg = QPrintPreviewDialog(self)
        ppdlg.paintRequested.connect(self.centralW.view.printView)
        ppdlg.exec()

    def imageExportClick(self):
        image = QImage(self.centralW.view.viewport().size(),
                       QImage.Format_ARGB32_Premultiplied)
        self.centralW.view.printView(image)
        fdlg = QFileDialog(self, caption="Select or create an image file")
        fdlg.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
        fdlg.setDefaultSuffix("png")
        fdlg.setFileMode(QFileDialog.FileMode.AnyFile)
        fdlg.setViewMode(QFileDialog.ViewMode.Detail)
        fdlg.setNameFilter("Image Files (*.png *.jpg *.bmp *.gif *.jpeg)")
        if fdlg.exec() == QDialog.DialogCode.Accepted:
            imageFile = fdlg.selectedFiles()[0]
            image.save(imageFile)

    def deleteClick(self, s):
        self.centralW.scene.editModes.setMode("deleteItem")
        self.centralW.scene.deleteSelectedItems()
        self.messageLine.setText(self.centralW.scene.messages[
                                     self.centralW.scene.editModes.mode()])

    def selectAllClick(self):
        self.centralW.scene.selectAll()

    def deselectAllClick(self):
        self.centralW.scene.deselectAll()

    def stretchClick(self, s):
        self.centralW.scene.editModes.setMode("stretchItem")
        self.messageLine.setText(self.centralW.scene.messages[
                                     self.centralW.scene.editModes.mode()])
        self.centralW.scene.stretchSelectedItems()

    def moveClick(self):
        self.centralW.scene.editModes.setMode("moveItem")
        self.messageLine.setText(self.centralW.scene.messages[
                                     self.centralW.scene.editModes.mode()])

    def moveByClick(self):
        self.centralW.scene.editModes.setMode("moveItem")
        self.messageLine.setText(self.centralW.scene.messages[
                                     self.centralW.scene.editModes.mode()])
        self.centralW.scene.moveBySelectedItems()

    def rotateClick(self):
        self.centralW.scene.editModes.setMode("rotateItem")
        self.messageLine.setText(self.centralW.scene.messages[
                                     self.centralW.scene.editModes.mode()])

    def panView(self):
        self.centralW.scene.editModes.setMode("panView")
        self.messageLine.setText(self.centralW.scene.messages[
                                     self.centralW.scene.editModes.mode()])

    def alignItemsClick(self):
        self.messageLine.setText("Select Items will be aligned.")
        dlg = alg.alignItemsDialogue(self)
        dlg.buttonBox.button(
            QDialogButtonBox.StandardButton.Apply).clicked.connect(
            lambda: alg.handleAlignAction(dlg, False))
        dlg.buttonBox.button(QDialogButtonBox.StandardButton.Ok).clicked.connect(
            lambda: alg.handleAlignAction(dlg, True))
        dlg.alignLineButton.pressed.connect(
            lambda: alg.startLineAlign(self.centralW.scene, dlg))
        dlg.alignEdgesButton.pressed.connect(
            lambda: alg.startEdgeAlign(self.centralW.scene, dlg))
        dlg.show()

    def goUpHierarchy(self):
        self.saveCell()
        if self.parentEditor is not None:
            self.parentEditor.updateDesignScene()
            self.parentEditor.raise_()
        self.close()

    def fitToWindow(self):
        self.centralW.scene.fitItemsInView()
        self.messageLine.setText("Fitting to window")

    def copyClick(self, s):
        self.centralW.scene.editModes.setMode("copyItem")
        self.messageLine.setText(self.centralW.scene.messages[
                                     self.centralW.scene.editModes.mode()])
        self.centralW.scene.copySelectedItems()

    def horizontalFlipClick(self):
        self.messageLine.setText('Flipping Selected Items Horizontally')
        self.centralW.scene.flipHorizontal()

    def verticalFlipClick(self):
        self.messageLine.setText('Flipping Selected Items Vertically')
        self.centralW.scene.flipVertical()

    def zoomIn(self):
        self.messageLine.setText("Zooming in")
        self.centralW.view.scale(1.25, 1.25)
        self.centralW.view.viewRect = self.centralW.view.mapToScene(
            self.rect()).boundingRect().toRect()

    def zoomOut(self):
        self.messageLine.setText("Zooming out")
        self.centralW.view.scale(0.8, 0.8)
        self.centralW.view.viewRect = self.centralW.view.mapToScene(
            self.rect()).boundingRect().toRect()

    def closeWindow(self):
        self.close()

    def closeEvent(self, event):
        cellViewTuple = ddef.viewNameTuple(self.libName, self.cellName, self.viewName)
        self.appMainW.openViews.pop(cellViewTuple, None)
        event.accept()
        super().closeEvent(event)

    def moveOrigin(self):
        self.centralW.scene.editModes.setMode("changeOrigin")

    def undoClick(self, s):

        self.messageLine.setText(self.centralW.scene.undoStack.undoText())
        self.centralW.scene.undoStack.undo()

    def redoClick(self, s):
        self.centralW.scene.undoStack.redo()

    def rotateItemClick(self, s):
        self.centralW.scene.editModes.setMode("rotateItem")

    def redraw(self):
        self.messageLine.setText("Redrawing...")
        self.centralW.view.updateSceneRect(self.centralW.view.sceneRect())

    def _findTopParent(self):
        current = self
        while current is not None and hasattr(current, 'parentObj'):
            if current.parentObj is None:
                return current
            current = current.parentObj
        return current

    def _createSignalConnections(self):
        self.centralW.scene.selectionChanged.connect(
            self.appMainW.selectionChangedScene)
        self.centralW.view.keyPressedSignal.connect(self.appMainW.viewKeyPressed)

    def addPluginAction(self, menu_name: str, action_name: str, callback, shortcut=None):
        """Add plugin action to specified menu"""
        menu = None
        for action in self.menuBar().actions():
            if action.text() == menu_name:
                menu = action.menu()
                break

        if not menu:
            menu = self.menuBar().addMenu(menu_name)

        action = QAction(action_name, self)
        action.triggered.connect(callback)
        if shortcut:
            action.setShortcut(shortcut)

        menu.addAction(action)
        return action

    @contextmanager
    def measureDuration(self):
        start_time = time.perf_counter()
        try:
            yield
        finally:
            end_time = time.perf_counter()
            self.logger.info(
                f"Total processing time: {(end_time - start_time) * 1000:.3f} milliseconds")


class editorContainer(QWidget):
    def __init__(self, parent: editorWindow):
        super().__init__(parent=parent)
        self.editorWindow = parent
        self.scene = editorScene(self)
        self.view = editorView(self.scene, self)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#    Add-ons and extensions developed for this software may be distributed
#    under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
import importlib
import os
import pathlib
import sys
from types import ModuleType
from typing import Union

from PySide6.QtGui import (QAction, QIcon)
from PySide6.QtWidgets import QApplication
from dotenv import load_dotenv

load_dotenv()

_module_cache = {}


def _get_pdk_path() -> pathlib.Path:
    """Get and cache PDK path"""
    base_path = pathlib.Path(__file__).resolve().parent.parent.parent
    pdkPath = os.environ.get("REVEDA_PDK_PATH")
    
    if pdkPath:
        pdkPathObj = pathlib.Path(pdkPath)
        if not pdkPathObj.is_absolute():
            pdkPathObj = base_path / pdkPath
    else:
        pdkPathObj = base_path / 'defaultPDK'
    
    pdkPathObj = pdkPathObj.resolve()

    if not pdkPathObj.exists():
        pdkPathObj = (base_path / 'defaultPDK').resolve()
        if not pdkPathObj.exists():
            raise FileNotFoundError(f"PDK path not found: {pdkPathObj}")

    pdkPathParentStr = str(pdkPathObj.parent)
    if pdkPathParentStr not in sys.path:
        sys.path.append(pdkPathParentStr)

    return pdkPathObj


def importPDKModule(moduleName) -> Union[ModuleType, None]:
    """Import a PDK submodule dynamically"""
    if moduleName in _module_cache:
        return _module_cache[moduleName]

    pdkPathObj = _get_pdk_path()
    fullModuleName = f"{pdkPathObj.name}.{moduleName}"

    try:
        module = importlib.import_module(fullModuleName)
        _module_cache[moduleName] = module
        return module
    except ModuleNotFoundError:
        _module_cache[moduleName] = None
        return None


class pdkConfig:
    def __init__(self, pdkPathObj):
        self_app = QApplication.instance()
        self.pdkPathObj = pdkPathObj
        self.pdkConfig = self._loadPDKConfig()

    def _loadPDKConfig(self):
        """Load PDK configuration from JSON file"""
        import json
        if not self.pdkPathObj.exists():
            return {}

        try:
            with open(self.pdkPathObj / "config.json") as f:
                config = json.load(f)
                return config
        except Exception as e:
            self._app.logger.warning(
                f"Failed to load plugin config for {item.name}: {e}")
            return {}

    def applyPDKMenus(self, editorWindow):
        if self.pdkConfig:
            editorClassName = editorWindow.__class__.__name__
            for menuItem in self.pdkConfig.get("menu_items", []):
                # Check applicability
                if "apply" in menuItem and editorClassName not in menuItem["apply"]:
                    continue
                # Get callback
                module = importPDKModule(menuItem["module"])
                if not module:
                    continue
                callback = getattr(module, menuItem["callback"], None)
                if not callback:
                    continue

                # Find target menu and add action
                if hasattr(editorWindow, menuItem['location']):
                    for action in editorWindow.menuBar().actions():
                        if action.text().replace('&', '') == menuItem["menu"]:
                            new_action = QAction(menuItem["action"], editorWindow)
                            if "text" in menuItem:
                                new_action.setText(menuItem["text"])
                            if "icon" in menuItem:
                                new_action.setIcon(QIcon(menuItem["icon"]))
                            if "shortcut" in menuItem:
                                new_action.setShortcut(menuItem["shortcut"])
                            if "checked" in menuItem:
                                new_action.setCheckable(True)
                                new_action.setChecked(menuItem["checked"])
                            new_action.triggered.connect(lambda c=False, cb=callback: cb(
                                editorWindow))
                            action.menu().addAction(new_action)
                            break


def getPDKPath():
    """Get the resolved PDK path"""
    return _get_pdk_path()



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

# Layout Layers
from PySide6.QtCore import Qt
from PySide6.QtGui import QColor

import revedaEditor.backend.dataDefinitions as ddef

odLayer_drw = ddef.layLayer(
    name="od",
    purpose="drw",
    pcolor=QColor(255, 0, 0, 127),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(255, 0, 0, 127),
    btexture="stipples/stipple1.txt",
    z=1,
    visible=True,
    selectable=True,
    gdsLayer=0,
)

odLayer_pin = ddef.layLayer(
    name="od",
    purpose="pin",
    pcolor=QColor(255, 0, 0, 180),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(139, 0, 0, 180),
    btexture="stipples/stipple1.txt",
    z=1,
    visible=True,
    selectable=True,
    gdsLayer=0,
)
odLayer_txt = ddef.layLayer(
    name="od",
    purpose="txt",
    pcolor=QColor(255, 0, 0, 255),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(139, 0, 0, 255),
    btexture="stipples/stipple1.txt",
    z=1,
    visible=True,
    selectable=True,
    gdsLayer=0,
)
activeLayer_drw = ddef.layLayer(
    name="active",
    purpose="drw",
    pcolor=QColor(0, 0, 255, 127),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(0, 0, 255, 180),
    btexture="stipples/stipple2.txt",
    z=2,
    visible=True,
    selectable=True,
    gdsLayer=1,
)
activeLayer_pin = ddef.layLayer(
    name="active",
    purpose="pin",
    pcolor=QColor(0, 0, 255, 180),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(0, 0, 139, 180),
    btexture="stipples/stipple2.txt",
    z=2,
    visible=True,
    selectable=True,
    gdsLayer=1,
)
activeLayer_txt = ddef.layLayer(
    name="active",
    purpose="txt",
    pcolor=QColor(0, 0, 255, 255),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(0, 0, 255, 255),
    btexture="stipples/stipple2.txt",
    z=2,
    visible=True,
    selectable=True,
    gdsLayer=1,
)
contactLayer_drw = ddef.layLayer(
    name="contact",
    purpose="drw",
    pcolor=QColor(0, 255, 0, 127),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor("green"),
    btexture="stipples/stipple3.txt",
    z=3,
    visible=True,
    selectable=True,
    gdsLayer=2,
)
contactLayer_pin = ddef.layLayer(
    name="contact",
    purpose="pin",
    pcolor=QColor(0, 0, 255, 180),
    pwidth=3,
    pstyle=Qt.SolidLine,
    bcolor=QColor("darkGreen"),
    btexture="stipples/stipple3.txt",
    z=3,
    visible=True,
    selectable=True,
    gdsLayer=2,
)
contactLayer_txt = ddef.layLayer(
    name="contact",
    purpose="txt",
    pcolor=QColor(50, 0, 255, 255),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(50, 0, 255, 255),
    btexture="stipples/stipple3.txt",
    z=3,
    visible=True,
    selectable=True,
    gdsLayer=2,
)
poLayer_drw = ddef.layLayer(
    name="po",
    purpose="drw",
    pcolor=QColor(42, 170, 138, 127),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(42, 170, 138, 127),
    btexture="stipples/stipple7.txt",
    z=4,
    visible=True,
    selectable=True,
    gdsLayer=3,
)
poLayer_pin = ddef.layLayer(
    name="po",
    purpose="pin",
    pcolor=QColor(42, 170, 138, 180),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(42, 170, 138, 180),
    btexture="stipples/stipple7.txt",
    z=4,
    visible=True,
    selectable=True,
    gdsLayer=3,
)
poLayer_txt = ddef.layLayer(
    name="po",
    purpose="txt",
    pcolor=QColor(42, 170, 138, 255),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(42, 170, 138, 255),
    btexture="stipples/stipple7.txt",
    z=4,
    visible=True,
    selectable=True,
    gdsLayer=3,
)
m1Layer_drw = ddef.layLayer(
    name="m1",
    purpose="drw",
    pcolor=QColor(255, 255, 0, 127),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(255, 255, 0, 127),
    btexture="stipples/stipple4.txt",
    z=4,
    visible=True,
    selectable=True,
    gdsLayer=3,
)
m1layer_pin = ddef.layLayer(
    name="m1",
    purpose="pin",
    pcolor=QColor(246, 190, 0, 180),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(246, 190, 0, 180),
    btexture="stipples/stipple4.txt",
    z=4,
    visible=True,
    selectable=True,
    gdsLayer=3,
)
m1layer_txt = ddef.layLayer(
    name="m1",
    purpose="txt",
    pcolor=QColor(246, 190, 0, 255),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(246, 190, 0, 255),
    btexture="stipples/stipple4.txt",
    z=4,
    visible=True,
    selectable=True,
    gdsLayer=3,
)
m2Layer_drw = ddef.layLayer(
    name="m2",
    purpose="drw",
    pcolor=QColor(255, 0, 255, 127),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(255, 0, 255, 127),
    btexture="stipples/stipple5.txt",
    z=5,
    visible=True,
    selectable=True,
    gdsLayer=4,
)
m2Layer_pin = ddef.layLayer(
    name="m2",
    purpose="pin",
    pcolor=QColor(255, 0, 255, 180),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(255, 0, 255, 180),
    btexture="stipples/stipple5.txt",
    z=5,
    visible=True,
    selectable=True,
    gdsLayer=4,
)
m2Layer_txt = ddef.layLayer(
    name="m2",
    purpose="txt",
    pcolor=QColor(255, 0, 255, 255),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(255, 0, 255, 255),
    btexture="stipples/stipple5.txt",
    z=5,
    visible=True,
    selectable=True,
    gdsLayer=4,
)
via1Layer_drw = ddef.layLayer(
    name="via1",
    purpose="drw",
    pcolor=QColor(180, 0, 180, 127),
    pwidth=2,
    pstyle=Qt.SolidLine,
    bcolor=QColor(180, 0, 180, 127),
    btexture="stipples/stipple6.txt",
    z=6,
    visible=True,
    selectable=True,
    gdsLayer=5,
)
via1Layer_pin = ddef.layLayer(
    name="via1",
    purpose="pin",
    pcolor=QColor(180, 0, 180, 180),
    pwidth=3,
    pstyle=Qt.SolidLine,
    bcolor=QColor(180, 0, 180, 180),
    btexture="stipples/stipple6.txt",
    z=6,
    visible=True,
    selectable=True,
    gdsLayer=5,
)
via1Layer_txt = ddef.layLayer(
    name="via1",
    purpose="txt",
    pcolor=QColor(180, 0, 180, 255),
    pwidth=1,
    pstyle=Qt.SolidLine,
    bcolor=QColor(180, 0, 180, 255),
    btexture="stipples/stipple6.txt",
    z=6,
    visible=True,
    selectable=True,
    gdsLayer=5,
)
pdkDrawingLayers = [
    odLayer_drw,
    poLayer_drw,
    activeLayer_drw,
    contactLayer_drw,
    m1Layer_drw,
    m2Layer_drw,
    via1Layer_drw,
]
pdkPinLayers = [
    odLayer_pin,
    poLayer_pin,
    activeLayer_pin,
    contactLayer_pin,
    m1layer_pin,
    m2Layer_pin,
]
pdkViaLayers = [
    contactLayer_drw,
    via1Layer_drw,
]
pdkTextLayers = [
    odLayer_txt,
    poLayer_txt,
    activeLayer_txt,
    contactLayer_txt,
    m1layer_txt,
    m2Layer_txt,
]
pdkAllLayers = pdkDrawingLayers + pdkPinLayers + pdkTextLayers + pdkViaLayers



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

# Load symbol and maybe later schematic from json file.
# import pathlib

import functools
import json
import pathlib
from typing import Any, List

import orjson
from PySide6.QtCore import QPoint, QLineF, QRect
from PySide6.QtWidgets import (
    QGraphicsScene,
    QGraphicsRectItem,
    QGraphicsSimpleTextItem,
)

import revedaEditor.common.labels as lbl
import revedaEditor.common.layoutShapes as lshp
import revedaEditor.common.net as net
import revedaEditor.common.shapes as shp
import revedaEditor.fileio.symbolEncoder as se
from revedaEditor.backend.pdkLoader import importPDKModule

laylyr = importPDKModule('layoutLayers')
pcells = importPDKModule('pcells')
fabproc = importPDKModule('process')


class symbolItems:
    def __init__(self, scene: QGraphicsScene):
        """
        Initializes the class instance.

        Args:
            scene (QGraphicsScene): The QGraphicsScene object.

        """
        self.scene = scene
        self.snapTuple = scene.snapTuple

    def create(self, item: dict):
        """
        Create symbol items from json file.
        """
        if isinstance(item, dict):
            match item.get("type"):
                case "rect":
                    return self.createRectItem(item)
                case "circle":
                    return self.createCircleItem(item)
                case "arc":
                    return self.createArcItem(item)
                case "line":
                    return self.createLineItem(item)
                case "pin":
                    return self.createPinItem(item)
                case "label":
                    return self.createLabelItem(item)
                case "text":
                    return self.createTextItem(item)
                case "polygon":
                    return self.createPolygonItem(item)

    @staticmethod
    def createRectItem(item: dict):
        """
        Create symbol items from json file.
        """
        start = QPoint(item["rect"][0], item["rect"][1])
        end = QPoint(item["rect"][2], item["rect"][3])
        rect = shp.symbolRectangle(start, end)
        rect.setPos(
            QPoint(item["loc"][0], item["loc"][1]),
        )
        rect.angle = item.get("ang", 0)
        rect.flipTuple = item.get('fl', (1, 1))
        return rect

    @staticmethod
    def createCircleItem(item: dict):
        centre = QPoint(item["cen"][0], item["cen"][1])
        end = QPoint(item["end"][0], item["end"][1])
        circle = shp.symbolCircle(centre, end)  # note that we are using grid
        # values for
        # scene
        circle.setPos(
            QPoint(item["loc"][0], item["loc"][1]),
        )
        circle.angle = item.get("ang", 0)
        circle.flipTuple = item.get('fl', (1, 1))
        return circle

    @staticmethod
    def createArcItem(item: dict):
        start = QPoint(item["st"][0], item["st"][1])
        end = QPoint(item["end"][0], item["end"][1])

        arc = shp.symbolArc(start, end)  # note that we are using grid values
        # for scene
        arc.setPos(QPoint(item["loc"][0], item["loc"][1]))
        arc.angle = item.get("ang", 0)
        arc.flipTuple = item.get('fl', (1, 1))
        arc.arcType = shp.symbolArc.arcTypes[item["at"]]
        return arc

    @staticmethod
    def createLineItem(item: dict):
        start = QPoint(item["st"][0], item["st"][1])
        end = QPoint(item["end"][0], item["end"][1])

        line = shp.symbolLine(start, end)
        line.setPos(QPoint(item["loc"][0], item["loc"][1]))
        line.angle = item.get("ang", 0)
        line.flipTuple = item.get('fl', (1, 1))
        return line

    @staticmethod
    def createPinItem(item: dict):
        start = QPoint(item["st"][0], item["st"][1])
        pin = shp.symbolPin(start, item["nam"], item["pd"], item["pt"])
        pin.setPos(QPoint(item["loc"][0], item["loc"][1]))
        pin.angle = item["ang"]
        pin.flipTuple = item.get('fl', (1, 1))
        return pin

    @staticmethod
    def createLabelItem(item: dict):
        start = QPoint(item["st"][0], item["st"][1])
        label = lbl.symbolLabel(
            start,
            item["def"],
            item["lt"],
            item["ht"],
            item["al"],
            item["or"],
            item["use"],
        )
        label.setPos(QPoint(item["loc"][0], item["loc"][1]))
        label.labelName = item["nam"]
        label.labelText = item["txt"]
        label.labelVisible = item["vis"]
        label.labelValue = item["val"]
        return label

    @staticmethod
    def createTextItem(item: dict):
        start = QPoint(item["st"][0], item["st"][1])
        text = shp.text(
            start,
            item["tc"],
            item["ff"],
            item["fs"],
            item["th"],
            item["ta"],
            item["to"],
        )
        text.setPos(QPoint(item["loc"][0], item["loc"][1]))
        return text

    @staticmethod
    def createPolygonItem(item: dict):
        pointsList = [QPoint(point[0], point[1]) for point in item["ps"]]
        polygon = shp.symbolPolygon(pointsList)
        polygon.flipTuple = item.get('fl', (1, 1))
        return polygon

    @staticmethod
    def createSymbolAttribute(item: dict):
        return se.symbolAttribute(item["nam"], item["def"])

    def createSimpleTextItem(self, item: dict):
        text = QGraphicsSimpleTextItem(item["text"])
        text.setPos(QPoint(item["pos"][0], item["pos"][1]))
        return text

    def createQRectItem(self, item: dict):
        rect = QGraphicsRectItem(QRect(*item["rect"]))
        rect.setPos(QPoint(item["pos"][0], item["pos"][1]))
        return rect

    def unknownItem(self):
        rectItem = QGraphicsRectItem(QRect(0, 0, *self.snapTuple))
        rectItem.setVisible(False)
        return rectItem


class schematicItems:
    def __init__(self, scene: QGraphicsScene):
        self.scene = scene
        self.libraryDict = scene.libraryDict
        self.snapTuple = scene.snapTuple

    @staticmethod
    @functools.lru_cache(maxsize=256)
    def _load_sym_json(file_path_str: str):
        """Cache symbol file contents keyed by path to avoid repeated I/O for repeated cells."""
        try:
            with open(file_path_str, "rb") as f:
                return orjson.loads(f.read())
        except (orjson.JSONDecodeError, FileNotFoundError, OSError):
            return None

    def create(self, item: dict):
        if isinstance(item, dict):
            match item["type"]:
                case "sys":
                    return self._createSymbolShape(item)
                case "scn":
                    return self._createNet(item)
                case "scp":
                    return self._createPin(item)
                case "txt":
                    return self._createText(item)
                case _:
                    pass
                    # return self.unknownItem()

    def _createText(self, item):
        start = QPoint(0, 0)
        text = shp.text(
            start,
            item["tc"],
            item["ff"],
            item["fs"],
            item["th"],
            item["ta"],
            item["to"],
        )
        text.setPos(QPoint(item["st"][0], item["st"][1]))
        text.flipTuple = item.get('fl', (1, 1))
        text.angle = item.get('ang', 0)

        return text

    def _createPin(self, item):
        start = QPoint(0, 0)
        pinName = item["pn"]
        pinDir = item["pd"]
        pinType = item["pt"]
        pinItem = shp.schematicPin(
            start,
            pinName,
            pinDir,
            pinType,
        )
        pinItem.setPos(QPoint(item["st"][0], item["st"][1]))
        pinItem.angle = item.get('ang', 0)
        pinItem.flipTuple = item.get('fl', (1, 1))
        return pinItem

    def _createNet(self, item):
        start = QPoint(item["st"][0], item["st"][1])
        end = QPoint(item["end"][0], item["end"][1])
        width = item.get('w', 0)
        netItem = net.schematicNet(start, end, width)
        netItem.name = item["nam"]
        match item["ns"]:
            case 3:

                netItem.nameStrength = net.netNameStrengthEnum.SET
            case 2:

                netItem.nameStrength = net.netNameStrengthEnum.INHERIT
            case 1:

                netItem.nameStrength = net.netNameStrengthEnum.WEAK
            case _:
                netItem.nameStrength = net.netNameStrengthEnum.NONAME

        return netItem

    def _createSymbolShape(self, item):
        itemShapes = list()
        symbolAttributes = dict()
        symbolInstance = shp.schematicSymbol(itemShapes, symbolAttributes)
        symbolInstance.libraryName = item["lib"]
        symbolInstance.cellName = item["cell"]
        symbolInstance.viewName = item["view"]
        symbolInstance.counter = item["ic"]
        symbolInstance.instanceName = item["nam"]
        symbolInstance.netlistIgnore = bool(item.get("ign", 0))
        labelDict = item["ld"]
        symbolInstance.setPos(*item["loc"])
        [
            labelItem.labelDefs()
            for labelItem in symbolInstance.labels.values()
        ]
        libraryPath = self.libraryDict.get(item["lib"])
        if libraryPath is None:
            self.createDraftSymbol(item, symbolInstance)
            self.scene.logger.warning(f"{item['lib']} cannot be found.")
            return symbolInstance
        else:
            # find the symbol file
            file = libraryPath.joinpath(
                item["cell"], f'{item["view"]}.json'
            )
            if not file.exists():
                self.createDraftSymbol(item, symbolInstance)
                self.scene.logger.warning(f"{item['lib']} cannot be found.")
                return symbolInstance
            else:
                # load json file and create shapes
                file_path_str = str(file)
                jsonItems = self._load_sym_json(file_path_str)
                if jsonItems is None:
                    self.scene.logger.error("Error: Invalid or missing Symbol file")
                    return None
                try:
                    symbolShape = symbolItems(self.scene)
                    for jsonItem in jsonItems[2:]:  # skip first two entries.
                        if jsonItem["type"] == "attr":
                            symbolAttributes[jsonItem["nam"]] = (
                                jsonItem["def"]
                            )
                        else:
                            itemShapes.append(
                                symbolShape.create(jsonItem)
                            )
                    symbolInstance.shapes = itemShapes
                    for labelItem in symbolInstance.labels.values():
                        if (
                                labelItem.labelName
                                in labelDict.keys()
                        ):
                            labelItem.labelValue = (
                                labelDict[
                                    labelItem.labelName
                                ][0]
                            )
                            labelItem.labelVisible = (
                                labelDict[
                                    labelItem.labelName
                                ][1]
                            )
                    symbolInstance.symattrs = symbolAttributes
                    [
                        labelItem.labelDefs()
                        for labelItem in symbolInstance.labels.values()
                    ]
                    symbolInstance.angle = item.get("ang", 0)
                    symbolInstance.flipTuple = item.get('fl', (1, 1))
                    return symbolInstance
                except Exception as e:
                    self.scene.logger.error(
                        f"Error creating symbol instance: {e}"
                    )
                    return None

    def createDraftSymbol(self, item: dict, symbolInstance: shp.schematicSymbol):
        rectItem = shp.symbolRectangle(
            QPoint(item["br"][0], item["br"][1]), QPoint(item["br"][2], item["br"][3])
        )
        fixedFont = self.scene.fixedFont
        textItem = shp.text(
            rectItem.start,
            f'{item["lib"]}/{item["cell"]}/{item["view"]}',
            fixedFont.family(),
            fixedFont.styleName(),
            fixedFont.pointSize(),
            shp.text.textAlignments[0],
            shp.text.textOrients[0],
        )
        symbolInstance.shapes = [rectItem, textItem]
        symbolInstance.draft = True

    def unknownItem(self):
        rectItem = QGraphicsRectItem(QRect(0, 0, *self.snapTuple))
        rectItem.setVisible(False)
        return rectItem


class PCellCache:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(PCellCache, cls).__new__(cls)
            cls._instance.layout_file_cache = {}
        return cls._instance

    @classmethod
    @functools.lru_cache(maxsize=100)
    def getPCellDef(cls, file_path: str) -> dict:
        try:
            with open(file_path, "r") as temp:
                return json.load(temp)
        except (json.JSONDecodeError, FileNotFoundError):
            return {}

    @classmethod
    @functools.lru_cache(maxsize=100)
    def getPCellClass(cls, pcell_class_name: str) -> Any:
        return pcells.pcells.get(pcell_class_name)

    @classmethod
    def getLayoutFileContents(cls, file_path: str) -> List:
        return cls._instance.layout_file_cache.get(file_path)

    @classmethod
    def setLayoutFileContents(cls, file_path: str, contents: List):
        cls._instance.layout_file_cache[file_path] = contents

    @classmethod
    def clear_caches(cls):
        cls.getPCellDef.cache_clear()
        cls.getPCellClass.cache_clear()
        cls._instance.layout_file_cache.clear()


class layoutItems:
    def __init__(self, scene):
        self.scene = scene
        self.libraryDict = scene.libraryDict
        self.rulerFont = scene.rulerFont
        self.rulerTickLength = scene.rulerTickLength
        self.snapTuple = scene.snapTuple
        self.rulerWidth = scene.rulerWidth
        self.rulerTickGap = scene.rulerTickGap
        self.cache = PCellCache()

        # Pre-create method mapping for faster dispatch
        self._creators = {
            "Inst": self.createLayoutInstance,
            "Pcell": self.createPcellInstance,
            "Rect": self.createRectShape,
            "Path": self.createPathShape,
            "Label": self.createLabelShape,
            "Pin": self.createPinShape,
            "Polygon": self.createPolygonShape,
            "Via": self.createViaArrayShape,
            "Ruler": self.createRulerShape,
        }

    def create(self, item):
        if not isinstance(item, dict):
            return self.unknownItem()
        return self._creators.get(item.get("type"), self.unknownItem)(item)

    def _get_library_path(self, lib_name):
        """Common method to get and validate library path"""
        return self._get_library_path_cached(lib_name, tuple(self.libraryDict.items()))

    @functools.lru_cache(maxsize=32)
    def _get_library_path_cached(self, lib_name, library_dict_items):
        """Cached library path lookup"""
        library_dict = dict(library_dict_items)
        library_path = pathlib.Path(library_dict.get(lib_name, ""))
        if not library_path.exists():
            return None
        return library_path

    @staticmethod
    @functools.lru_cache(maxsize=128)
    def _load_json_file(file_path_str):
        """Cached JSON file loading with error handling"""
        try:
            with open(file_path_str, "rb") as f:
                return orjson.loads(f.read())
        except (orjson.JSONDecodeError, FileNotFoundError):
            return None

    def _set_common_attrs(self, obj, item):
        """Set common attributes for layout objects"""
        obj.angle = item.get("ang", 0)
        obj.flipTuple = item.get('fl', (1, 1))

    def createPcellInstance(self, item):
        library_path = self._get_library_path(item["lib"])
        if not library_path:
            return None

        file_path = library_path / item["cell"] / f"{item['view']}.json"
        pcell_def = self._load_json_file(str(file_path))
        if not pcell_def or pcell_def[0].get("cellView") != "pcell":
            self.scene.logger.error("Not a PCell cell")
            return None

        pcell_class = pcells.pcells.get(pcell_def[1].get("reference"))
        if not pcell_class:
            self.scene.logger.error(
                f"Unknown PCell class: {pcell_def[1].get('reference')}")
            return None

        try:
            instance = pcell_class()
            instance(**item.get("params", {}))
            instance.libraryName = item["lib"]
            instance.cellName = item["cell"]
            instance.viewName = item["view"]
            instance.counter = item["ic"]
            instance.instanceName = item["nam"]
            instance.setPos(QPoint(*item["loc"]))
            self._set_common_attrs(instance, item)
            return instance
        except Exception as e:
            self.scene.logger.error(f"Error creating PCell instance: {e}")
            return None

    def createLayoutInstance(self, item):
        library_path = self._get_library_path(item["lib"])
        if not library_path:
            return None

        file_path = library_path / item["cell"] / f"{item['view']}.json"
        file_path_str = str(file_path)

        # Use cache
        file_contents = self.cache.getLayoutFileContents(file_path_str)
        if file_contents is None:
            file_contents = self._load_json_file(file_path_str)
            if file_contents is None:
                return None
            self.cache.setLayoutFileContents(file_path_str, file_contents)

        # Create shapes with cached factory and pre-allocated list
        shapes_data = file_contents[2:]
        item_shapes = []
        item_shapes_append = item_shapes.append  # Cache method

        for shape in shapes_data:
            try:
                created_shape = self.create(
                    shape)  # Reuse self instead of creating new instance
                if created_shape:
                    item_shapes_append(created_shape)
            except Exception:
                pass  # Skip logging for performance

        instance = lshp.layoutInstance(item_shapes)
        loc = item["loc"]
        instance.libraryName = item["lib"]
        instance.cellName = item["cell"]
        instance.counter = item.get("ic")
        instance.instanceName = item.get("nam", "")
        instance.setPos(loc[0], loc[1])  # Cache loc lookup
        instance.viewName = item["view"]
        self._set_common_attrs(instance, item)
        return instance

    def createRectShape(self, item):
        tl, br, ln = item["tl"], item["br"], item["ln"]
        rect = lshp.layoutRect(
            QPoint(tl[0], tl[1]),
            QPoint(br[0], br[1]),
            laylyr.pdkAllLayers[ln]
        )
        self._set_common_attrs(rect, item)
        return rect

    def createPathShape(self, item):
        dfl1, dfl2 = item["dfl1"], item["dfl2"]
        path = lshp.layoutPath(
            QLineF(QPoint(dfl1[0], dfl1[1]), QPoint(dfl2[0], dfl2[1])),
            laylyr.pdkAllLayers[item["ln"]],
            item["w"], item["se"], item["ee"], item["md"]
        )
        path.name = item.get("nam", "")
        self._set_common_attrs(path, item)
        return path

    def createRulerShape(self, item):
        ruler = lshp.layoutRuler(
            QLineF(QPoint(*item["dfl1"]), QPoint(*item["dfl2"])),
            self.rulerWidth, self.rulerTickGap, self.rulerTickLength,
            self.rulerFont, item["md"]
        )
        ruler.angle = item.get("ang", 0)
        return ruler

    def createLabelShape(self, item):
        label = lshp.layoutLabel(
            QPoint(*item["st"]), item["lt"], item["ff"], item["fs"],
            item["fh"], item["la"], item["lo"], laylyr.pdkAllLayers[item["ln"]]
        )
        self._set_common_attrs(label, item)
        return label

    def createPinShape(self, item):
        pin = lshp.layoutPin(
            QPoint(*item["tl"]), QPoint(*item["br"]), item["pn"],
            item["pd"], item["pt"], laylyr.pdkAllLayers[item["ln"]]
        )
        self._set_common_attrs(pin, item)
        return pin

    @functools.lru_cache(maxsize=64)
    def _create_polygon_points(self, points_tuple):
        """Cache polygon point creation for repeated patterns"""
        return [QPoint(p[0], p[1]) for p in points_tuple]

    def createPolygonShape(self, item):
        points = self._create_polygon_points(tuple(tuple(p) for p in item["ps"]))
        polygon = lshp.layoutPolygon(points, laylyr.pdkAllLayers[item["ln"]])
        self._set_common_attrs(polygon, item)
        return polygon

    @functools.lru_cache(maxsize=16)
    def _get_via_def(self, via_name):
        """Cache via definition lookup"""
        return fabproc.processVias[fabproc.processViaNames.index(via_name)]

    def createViaArrayShape(self, item):
        via_info = item["via"]
        via_def = self._get_via_def(via_info["vdt"])
        via_st = via_info["st"]
        via = lshp.layoutVia(
            QPoint(via_st[0], via_st[1]), via_def,
            via_info["w"], via_info["h"]
        )
        st = item["st"]
        via_array = lshp.layoutViaArray(
            QPoint(st[0], st[1]), via,
            item["xs"], item["ys"], item["xn"], item["yn"]
        )
        self._set_common_attrs(via_array, item)
        return via_array

    def unknownItem(self):
        rect_item = QGraphicsRectItem(QRect(0, 0, *self.snapTuple))
        rect_item.setVisible(False)
        return rect_item



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import logging
import re
import xml.etree.ElementTree as ET
from pathlib import Path

from PySide6.QtCore import Qt
from PySide6.QtGui import (
    QColor,
)

from revedaEditor.backend.dataDefinitions import layLayer
from revedaEditor.gui.stippleEditor import stippleEditor

# Set up logging for better error reporting
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def parse_source_field(source_text: str) -> tuple[str, str, int, int]:
    """
    Parse the source field to extract name, purpose, gdsLayer, and dataType.
    
    Expected formats:
    - "COMP 22/0@1" -> name="COMP", purpose="drawing", gdsLayer=22, dataType=0
    - "pass_mk 2/222@1" -> name="pass_mk", purpose="drawing", gdsLayer=2, dataType=222
    - "5/222@1" -> name="layer5", purpose="drawing", gdsLayer=5, dataType=222
    
    Args:
        source_text: The source field text from the LYP file
        
    Returns:
        tuple: (name, purpose, gdsLayer, dataType)
    """
    if not source_text or not source_text.strip():
        logger.warning("Empty source field, using defaults")
        return "unknown", "drawing", 0, 0

    # Remove @ and everything after it
    clean_source = source_text.split('@')[0].strip()

    # Split on whitespace to separate potential name from layer/datatype
    parts = clean_source.split()

    if len(parts) == 1:
        # Format: "22/0" or just "22"
        layer_part = parts[0]
        name = None
    else:
        # Format: "COMP 22/0" or "pass_mk 2/222"
        # Everything except the last part is the name
        name = "_".join(parts[:-1])
        layer_part = parts[-1]

    # Parse the layer/datatype part
    if '/' in layer_part:
        try:
            gds_layer_str, datatype_str = layer_part.split('/', 1)
            gds_layer = int(float(gds_layer_str))
            datatype = int(float(datatype_str))
        except (ValueError, IndexError) as e:
            logger.warning(f"Failed to parse layer/datatype from '{layer_part}': {e}")
            gds_layer, datatype = 0, 0
    else:
        # Only layer number provided
        try:
            gds_layer = int(float(layer_part))
            datatype = 0
        except ValueError as e:
            logger.warning(f"Failed to parse layer from '{layer_part}': {e}")
            gds_layer, datatype = 0, 0

    # Generate name if not found
    if not name:
        name = f"layer{gds_layer}"

    # Use default purpose
    purpose = "drawing"

    return name, purpose, gds_layer, datatype


def safe_get_text(element, default: str = "") -> str:
    """Safely get text from an XML element, returning default if element is None or has no text."""
    if element is None:
        return default
    return element.text if element.text is not None else default


def safe_get_color(color_text: str, default_color: QColor = QColor(Qt.black)) -> QColor:
    """Safely parse color text, returning default color if parsing fails."""
    try:
        if color_text and color_text.strip():
            return QColor.fromString(color_text.upper())
    except Exception as e:
        logger.warning(f"Failed to parse color '{color_text}': {e}")
    return default_color


def safe_get_bool(bool_text: str, default: bool = True) -> bool:
    """Safely parse boolean text."""
    if not bool_text:
        return default
    return bool_text.lower() == 'true'


def safe_get_int(int_text: str, default: int = 1) -> int:
    """Safely parse integer text."""
    try:
        if int_text and int_text.strip():
            return int(float(int_text))
    except (ValueError, TypeError) as e:
        logger.warning(f"Failed to parse integer '{int_text}': {e}")
    return default


def parseLyp(lypFile: str, outputFileDir: str) -> bool:
    """
    Parse a LYP (Layer Properties) file and generate layoutLayers.py file.
    
    This function is more robust and handles cases where name fields are empty
    by parsing the source field to extract layer information.
    
    Args:
        lypFile: Path to the input LYP file
        outputFileDir: Directory where layoutLayers.py will be generated
        
    Returns:
        bool: True if parsing was successful, False otherwise
    """
    try:
        lypFileObj = Path(lypFile)
        outputFileDirObj = Path(outputFileDir)

        # Validate input file exists
        if not lypFileObj.exists():
            logger.error(f"LYP file does not exist: {lypFile}")
            return False

        # Create output directory if it doesn't exist
        outputFileDirObj.mkdir(parents=True, exist_ok=True)

        # Parse XML file
        try:
            tree = ET.parse(lypFileObj)
            root = tree.getroot()
        except ET.ParseError as e:
            logger.error(f"Failed to parse XML file '{lypFile}': {e}")
            return False

        purposeDict = {}
        layout_layers = []
        skipped_layers = 0

        with outputFileDirObj.joinpath("layoutLayers.py").open("w",
                                                               encoding="utf-8") as file:
            # Write header
            file.write("# Auto-generated layoutLayers.py from LYP file\n")
            file.write("# Do not edit this file manually\n\n")
            file.write("import revedaEditor.backend.dataDefinitions as ddef\n\n")

            logger.info(f"Processing LYP file: {lypFile}")

            for i, layerItem in enumerate(root.iterfind("properties")):
                try:
                    # Extract basic properties with safe parsing
                    frame_color_elem = layerItem.find("frame-color")
                    fill_color_elem = layerItem.find("fill-color")
                    dither_pattern_elem = layerItem.find("dither-pattern")
                    valid_elem = layerItem.find("valid")
                    visible_elem = layerItem.find("visible")
                    width_elem = layerItem.find("width")
                    name_elem = layerItem.find("name")
                    source_elem = layerItem.find("source")

                    # Parse colors
                    pcolor = safe_get_color(safe_get_text(frame_color_elem),
                                            QColor(Qt.black))
                    bcolor = safe_get_color(safe_get_text(fill_color_elem),
                                            QColor(Qt.transparent))

                    # Parse other properties
                    btexture = safe_get_text(dither_pattern_elem, "I3") + ".txt"
                    selectable = safe_get_bool(safe_get_text(valid_elem), True)
                    visible = safe_get_bool(safe_get_text(visible_elem), True)
                    pwidth = safe_get_int(safe_get_text(width_elem), 1)

                    # Get name and source text
                    name_text = safe_get_text(name_elem)
                    source_text = safe_get_text(source_elem)

                    # Determine layer name and properties
                    if name_text and name_text.strip():
                        # Use existing name if available
                        if "." in name_text:
                            name, purpose = name_text.split(".", 1)
                        else:
                            name = name_text
                            purpose = "drawing"

                        # Still need to parse source for GDS layer info
                        _, _, gdsLayer, dataType = parse_source_field(source_text)
                    else:
                        # Parse source field when name is empty
                        name, purpose, gdsLayer, dataType = parse_source_field(source_text)

                    # Validate required fields
                    if not name or not name.strip():
                        logger.warning(f"Skipping layer {i}: Invalid name")
                        skipped_layers += 1
                        continue

                    # Clean up name (remove invalid characters)
                    name = re.sub(r'[^\w]', '_', name.strip())
                    purpose = re.sub(r'[^\w]', '_', purpose.strip())

                    # Create layer object
                    layoutLayerItem = layLayer(
                        name=name,
                        purpose=purpose,
                        pcolor=pcolor,
                        pwidth=pwidth,
                        pstyle=Qt.SolidLine,
                        bcolor=bcolor,
                        btexture=btexture,
                        z=i,
                        selectable=selectable,
                        visible=visible,
                        gdsLayer=gdsLayer,
                        datatype=dataType,
                    )

                    # Add to collections
                    layout_layers.append(layoutLayerItem)
                    if purpose not in purposeDict:
                        purposeDict[purpose] = []
                    purposeDict[purpose].append(layoutLayerItem)

                    # Write layer definition
                    file.write(
                        f"{layoutLayerItem.name}_{layoutLayerItem.purpose} = "
                        f"ddef.{layoutLayerItem}\n"
                    )

                except Exception as e:
                    logger.error(f"Error processing layer {i}: {e}")
                    skipped_layers += 1
                    continue

            # Write purpose-based layer lists
            file.write("\n# Purpose-based layer lists\n")
            pdkAllLayerString = "pdkAllLayers = ["

            for purpose, layerList in purposeDict.items():
                if layerList:  # Only write non-empty lists
                    purposeString = f"pdk{purpose.capitalize()}Layers = ["
                    for layer in layerList:
                        layer_var = f"{layer.name}_{layer.purpose}"
                        purposeString += f" {layer_var},"
                        pdkAllLayerString += f" {layer_var},"
                    purposeString += " ]\n"
                    file.write(purposeString)

            pdkAllLayerString += " ]\n"
            file.write(pdkAllLayerString)

            logger.info(f"Successfully processed {len(layout_layers)} layers")
            if skipped_layers > 0:
                logger.warning(f"Skipped {skipped_layers} layers due to errors")

        # Process dither patterns
        process_dither_patterns(root, lypFileObj)

        return True

    except Exception as e:
        logger.error(f"Fatal error processing LYP file '{lypFile}': {e}")
        return False


def process_dither_patterns(root, lypFileObj: Path) -> None:
    """Process custom dither patterns from the LYP file."""
    try:
        for ditherPattern in root.iterfind("custom-dither-pattern"):
            for orderItem in ditherPattern.iterfind("order"):
                fileName = f"C{orderItem.text}.txt"
                fileObj = lypFileObj.parent.joinpath(fileName)

                with fileObj.open("w", encoding="utf-8") as patternFile:
                    for pattern in ditherPattern.findall("pattern"):
                        for lineItem in pattern.iterfind("line"):
                            if lineItem.text:
                                patternFile.write(
                                    f"{lineItem.text.replace('.', '0 ').replace('*', '1 ')}\n"
                                )

                # Generate PNG file
                imageFileObj = fileObj.with_suffix(".png")
                try:
                    stippleEditW = stippleEditor(None)
                    stippleEditW.loadPatternFromFile(str(fileObj))
                    stippleEditW.imageExportToFile(str(imageFileObj))
                except Exception as e:
                    logger.warning(f"Failed to generate PNG for pattern {fileName}: {e}")

    except Exception as e:
        logger.error(f"Error processing dither patterns: {e}")



from PySide6.QtCore import QPoint

from revedaEditor.common.labels import symbolLabel


class DummyParent:
    def __init__(self):
        self.cellName = "cell"
        self.instanceName = "I1"
        self.libraryName = "lib"
        self.viewName = "view"
        self.attr = {"modelName": "model"}
        self.counter = 42
        self.labels = {}


def make_label():
    label = symbolLabel(QPoint(0, 0), "[@testLabel:%]", "NLPLabel", 10, "Left",
                        "R0", "Normal")
    label.parentItem = lambda: DummyParent()
    return label


def test_basic_format_with_value():
    label = make_label()
    result = label.createNLPLabel("[@foo:%]", "bar")
    assert result == ("bar", "bar", "foo")


def test_basic_format_without_value():
    label = make_label()
    result = label.createNLPLabel("[@foo:%]", "")
    assert result == ("?", "?", "foo")


def test_format_with_default_value():
    label = make_label()
    result = label.createNLPLabel("[@foo:%:default=abc]", "")
    assert result == ("abc", "abc", "foo")


def test_format_with_default_value_and_value():
    label = make_label()
    result = label.createNLPLabel("[@foo:%:default=abc]", "xyz")
    assert result == ("xyz", "xyz", "foo")


def test_format_without_percent():
    label = make_label()
    result = label.createNLPLabel("[@foo:bar]", "baz")
    assert result == ("bar", "baz", "foo")


def test_invalid_label_definition():
    label = make_label()
    result = label.createNLPLabel("foo", "bar")
    assert result == ("foo", "bar", "foo")


def test_predefined_label_cellName():
    label = make_label()
    result = label.createNLPLabel("[@cellName]", "")
    assert result == ("cell", "cell", "cellName")


def test_predefined_label_instName():
    label = make_label()
    result = label.createNLPLabel("[@instName]", "")
    assert result == ("I1", "I1", "instName")


def test_predefined_label_libName():
    label = make_label()
    result = label.createNLPLabel("[@libName]", "")
    assert result == ("lib", "lib", "libName")


def test_predefined_label_viewName():
    label = make_label()
    result = label.createNLPLabel("[@viewName]", "")
    assert result == ("view", "view", "viewName")


def test_predefined_label_modelName():
    label = make_label()
    result = label.createNLPLabel("[@modelName]", "")
    assert result == ("model", "model", "modelName")


def test_predefined_label_elementNum():
    label = make_label()
    result = label.createNLPLabel("[@elementNum]", "")
    assert result == ("42", "42", "elementNum")



#!/usr/bin/env python3
"""
KLayout DRC XML report to dictionary converter using lxml.
Handles nested categories, cells, and DRC violations with polygons.
"""

import sys
from typing import Dict, Any, List

from PySide6.QtCore import (
    QPoint,
    Qt,
)
from PySide6.QtGui import QPolygonF, QPen, QColor
from PySide6.QtWidgets import (
    QApplication,
)
from PySide6.QtWidgets import (
    QGraphicsPolygonItem,
)
from lxml import etree

from revedaEditor.backend.pdkLoader import importPDKModule

process = importPDKModule("process")


class DRCErrorPolygon(QGraphicsPolygonItem):
    def __init__(self, polygon: QPolygonF) -> None:
        super().__init__(polygon)
        # self.setBrush(QBrush(QColor(255, 0, 0, 100)))
        self.setZValue(100)
        self.setPen(QPen(QColor(255, 0, 0), 20, Qt.DashLine))
        self._errorCategory = ""
        self._cell = ""

    def __repr__(self) -> str:
        return f"DRCErrorPolygon({self.polygon})"

    def __str__(self) -> str:
        return f"DRCErrorPolygon({self.polygon})"

    @property
    def errorCategory(self):
        return self._errorCategory

    @errorCategory.setter
    def errorCategory(self, value: str):
        if isinstance(value, str):
            self._errorCategory = value

    @property
    def cell(self):
        return self._cell

    @cell.setter
    def cell(self, value: str):
        if isinstance(value, str):
            self._cell = value


class DRCOutput:
    def __init__(self, path: str) -> None:
        self.path = path
        self.tree = etree.parse(path)
        self.root = self.tree.getroot()
        self.result = {}

    def parseDRCOutput(self) -> Dict[str, Any]:
        result = {
            "metadata": {},
            "categories": {},
            "cells": {},
            "violations": [],  # list of dicts for DRC items
        }

        # Parse top-level metadata
        for child in self.root:
            tag = child.tag
            if tag == "description":
                result["metadata"]["description"] = child.text
            elif tag == "generator":
                result["metadata"]["generator"] = child.text
            elif tag == "top-cell":
                result["metadata"]["top_cell"] = child.text
            elif tag == "categories":
                result["categories"] = self.parseCategories(child)
            elif tag == "cells":
                result["cells"] = self.parseCells(child)
            elif tag == "items":
                result["violations"] = self.parseViolations(child)
        self.result = result

    def parseCategories(self, categoryElement: etree._Element) -> Dict[str, Dict]:
        """Parse nested categories structure."""
        cats = {}
        for cat in categoryElement.findall("category"):
            name = cat.find("name")
            desc = cat.find("description")
            cats[name.text if name is not None else ""] = (
                desc.text if desc is not None else ""
            )
        return cats

    def parseCells(self, cells_el: etree._Element) -> Dict[str, Dict]:
        """Parse cells section."""
        cells = {}
        for cell in cells_el.findall("cell"):
            name = cell.find("name")
            cells[name.text] = self.xmltoDict(cell) if name is not None else {}
        return cells

    def xmltoDict(self, element: etree._Element) -> Dict[str, Any]:
        """Recursively convert lxml Element to dictionary."""
        if element.text and element.text.strip():
            return {
                "tag": element.tag,
                "text": element.text.strip(),
                "tail": element.tail,
            }

        result = {"tag": element.tag}

        # Handle attributes
        if element.attrib:
            result["attrib"] = dict(element.attrib)

        # Handle text content
        if element.text and element.text.strip():
            result["text"] = element.text.strip()

        # Handle children
        children = {}
        for child in element:
            child_dict = self.xmltoDict(child)
            child_tag = child_dict["tag"]
            if child_tag in children:
                if not isinstance(children[child_tag], list):
                    children[child_tag] = [children[child_tag]]
                children[child_tag].append(child_dict)
            else:
                children[child_tag] = child_dict

        if children:
            result["children"] = children

        return result

    def parseViolations(self, items_el: etree._Element):
        def parsePolygon(valueStr: str) -> tuple[str, List[QPoint], List[str]]:
            def parseCoords(point_list: List[str]) -> List[QPoint]:
                coords = []
                for pt in point_list:
                    pt = pt.strip()
                    if "," in pt:
                        try:
                            x, y = map(float, pt.split(","))
                            coords.append(QPoint(x * process.dbu, y * process.dbu))
                        except ValueError:
                            continue
                return coords

            if valueStr.startswith("polygon:"):
                points = valueStr[9:].strip("()").split(";")
                return ("polygon", parseCoords(points), points)
            elif valueStr.startswith("edge-pair:"):
                edge_str = valueStr[10:].strip()
                points = []
                for pair in edge_str.split("/"):
                    points.extend(pair.strip("()").split(";"))
                return ("edge-pair", parseCoords(points), points)
            return ("", [], [])

        """Parse violations/items section."""
        violations = []
        for item in items_el.findall("item"):
            violation = {
                "category": (
                    item.find("category").text.strip("'")
                    if item.find("category") is not None
                    else None
                ),
                "cell": (
                    item.find("cell").text if item.find("cell") is not None else None
                ),
                "visited": (
                    item.find("visited").text == "true"
                    if item.find("visited") is not None
                    else False
                ),
                "multiplicity": (
                    int(item.find("multiplicity").text)
                    if item.find("multiplicity") is not None
                    else 1
                ),
                "polygons": [],
                "points": [],
            }

            # Parse values (polygons)
            values = item.find("values")
            if values is not None:
                for value_el in values.findall("value"):
                    errorType, polyCoords, points = parsePolygon(value_el.text)
                    if polyCoords:
                        violation["points"].append(points)
                        polygonItem = DRCErrorPolygon(QPolygonF(polyCoords))
                        polygonItem.cell = violation["cell"]
                        polygonItem.errorCategory = errorType
                        polygonItem.setToolTip(
                            f'{violation['cell']}, {violation['category']}, {violation["points"]}'
                        )
                        violation["polygons"].append(polygonItem)

            violations.append(violation)
        return violations


def main(file_path: str):
    """Main function to parse DRC file and print JSON-like dict."""
    drcDataObj = DRCOutput(file_path)
    drcDataObj.parseDRCOutput()
    # import json
    # Pretty print with indentation
    # print(json.dumps(drc_data, indent=2, default=str))


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: python drc_parser.py <drc_xml_file>")
        sys.exit(1)
    main(sys.argv[1])



import pytest
from PySide6.QtCore import QPoint
from PySide6.QtWidgets import (QGraphicsScene, QGraphicsView)

from revedaEditor.common.shapes import schematicSymbol, symbolPin


@pytest.fixture
def setup_shape(qtbot):
    pin1 = symbolPin(QPoint(0, 0), "pin1", "Input", "Signal")
    pin2 = symbolPin(QPoint(0, 10), "pin2", "Output", "Signal")
    pin3 = symbolPin(QPoint(0, 20), "pin3", "Input", "Signal")
    symattrs = {"pinOrder": "pin1, pin2, pin3"}
    shapes = [pin1, pin2, pin3]
    shape = schematicSymbol(shapes, symattrs)
    view = QGraphicsView()
    qtbot.addWidget(view)
    scene = QGraphicsScene()
    view.setScene(scene)
    scene.addItem(shape)
    return (shape, scene, pin1, pin2, pin3)


def test_pins_without_pin_order(setup_shape):
    shape, scene, pin1, pin2, pin3 = setup_shape
    shape.symattrs = {}
    assert shape.pins == {
        "pin1": pin1,
        "pin2": pin2,
        "pin3": pin3}


def test_pins_with_pin_order(setup_shape):
    shape, scene, pin1, pin2, pin3 = setup_shape
    shape.symattrs = {"pinOrder": "pin1, pin2, pin3"}
    assert shape.pins == {
        "pin1": pin1,
        "pin2": pin2,
        "pin3": pin3}


def test_pins_with_changed_order(setup_shape):
    shape, scene, pin1, pin2, pin3 = setup_shape
    shape.symattrs = {"pinOrder": "pin2, pin1, pin3"}
    assert shape.pins == {"pin2": pin2, "pin1": pin1, "pin3": pin3}


def test_pins_with_partial_order(setup_shape):
    shape, scene, pin1, pin2, pin3 = setup_shape
    shape.symattrs = {"pinOrder": "pin2, pin3"}
    assert shape.pins == {"pin2": pin2, "pin3": pin3}


def test_pins_with_invalid_order(setup_shape):
    shape, scene, pin1, pin2, pin3 = setup_shape
    shape.symattrs = {"pinOrder": "pin2, invalidPin, pin3"}
    assert shape.pins == {"pin2": pin2, "pin3": pin3}



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#    Add-ons and extensions developed for this software may be distributed
#    under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
# nuitka-project-if: {OS} == "Darwin":
#    nuitka-project: --standalone
#    nuitka-project: --macos-create-app-bundle
# The PySide6 plugin covers qt-plugins
# nuitka-project: --standalone
# nuitka-project: --deployment
# nuitka-project: --enable-plugin=pyside6
# nuitka-project: --include-data-dir=docs=docs
# nuitka-project: --include-package=revedaEditor
# nuitka-project: --include-package=markdown
# nuitka-project: --include-package=polars
# nuitka-project: --include-module=pydoc
# nuitka-project: --include-package=xml
# nuitka-project: --include-module=PySide6.QtWebEngineWidgets
# nuitka-project: --include-module=PySide6.QtOpenGL
# nuitka-project: --nofollow-import-to=unittest
# nuitka-project: --nofollow-import-to=pytest
# nuitka-project: --nofollow-import-to=revedasim
# nuitka-project: --nofollow-import-to=revedaPlot
# nuitka-project: --nofollow-import-to=plugins
# nuitka-project: --nofollow-import-to=defaultPDK
# nuitka-project: --include-package-data=defaultPDK
# nuitka-project-if: {OS} == "Windows":
#    nuitka-project: --output-dir=C:\Users\eskiye50\dist
# nuitka-project-if: {OS} == "Linux":
#    nuitka-project: --output-dir=/home/eskiyerli/dist
# nuitka-project: --product-name="Revolution EDA"
# nuitka-project: --product-version="0.8.10"
# nuitka-project: --company-name="Revolution EDA"
# nuitka-project: --file-description="Electronic Design Automation Software for Professional Custom IC Design Engineers"
# nuitka-project: --windows-icon-from-ico=revedaCoreLogo.ico
# nuitka-project: --copyright="Revolution Semiconductor (C) 2026"

import logging
import os
import platform
import sys
from contextlib import redirect_stderr, redirect_stdout
from pathlib import Path

from dotenv import load_dotenv
from PySide6.QtWidgets import QApplication

import revedaEditor.gui.pythonConsole as pcon
import revedaEditor.gui.revedaMain as rvm
from revedaEditor.backend.pdkLoader import pdkConfig
from revedaEditor.backend.pluginsLoader import pluginsLoader


class revedaApp(QApplication):
    """Revolution EDA application with plugin support and path management."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.basePath = Path(__file__).resolve().parent
        load_dotenv()
        self._setupLogger()

        self._setupPaths()
        self.appMainW = rvm.MainWindow()

    def _setupLogger(self):
        """Initialize application logger."""
        self.logger = logging.getLogger("reveda")
        logFilePath = self.basePath / "reveda.log"
        handler = logging.FileHandler(logFilePath)
        handler.setLevel(logging.INFO)
        handler.setFormatter(
            logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        )
        self.logger.addHandler(handler)

    def _setupPaths(self):
        pdkPath = os.environ.get("REVEDA_PDK_PATH")
        if pdkPath:
            path_obj = Path(pdkPath)
            self.revedaPdkPathObj = (
                path_obj if path_obj.is_absolute() else self.basePath / pdkPath
            ).resolve()
            if self.revedaPdkPathObj.exists():
                sys.path.append(str(self.revedaPdkPathObj))
            else:
                self.revedaPdkPathObj = self.basePath / "defaultPDK"
                if self.revedaPdkPathObj.exists():
                    sys.path.append(str(self.revedaPdkPathObj))
                else:
                    self.logger.error("Default PDK path cannot be found.")
        else:
            self.revedaPdkPathObj = self.basePath / "defaultPDK"
            if self.revedaPdkPathObj.exists():
                sys.path.append(str(self.revedaPdkPathObj))
        if self.revedaPdkPathObj.exists():
            self.pdkConfigObj = pdkConfig(self.revedaPdkPathObj)
        else:
            self.pdkConfigObj = None

        pluginPath = os.environ.get("REVEDA_PLUGIN_PATH")
        if pluginPath:
            path_obj = Path(pluginPath)
            self.revedaPluginPathObj = (
                path_obj if path_obj.is_absolute() else self.basePath / pluginPath
            ).resolve()
            if self.revedaPluginPathObj.exists():
                sys.path.append(str(self.revedaPluginPathObj))
                self.pluginsObj = pluginsLoader(self.revedaPluginPathObj)
        else:
            self.pluginsObj = None

    def updatePDKPath(self, newPath: Path):
        """Update PDK path and persist to .env file"""
        self.revedaPdkPathObj = newPath.resolve()

        # Update environment variable
        os.environ["REVEDA_PDK_PATH"] = str(self.revedaPdkPathObj)

        # Update sys.path
        if str(self.revedaPdkPathObj) not in sys.path:
            sys.path.append(str(self.revedaPdkPathObj))

        # Persist to .env file
        self.update_env_file("REVEDA_PDK_PATH", str(self.revedaPdkPathObj))

        self.logger.info(f"PDK path updated to: {self.revedaPdkPathObj}")

    def updatePluginsPath(self, newPath: str):
        """Update plugin path and persist to .env file"""
        if newPath:
            self.revedaPluginPathObj = Path(newPath).resolve()
            # Update environment variable
            os.environ["REVEDA_PLUGIN_PATH"] = str(self.revedaPluginPathObj)
            self.pluginsObj = pluginsLoader(self.revedaPluginPathObj)
            # Update sys.path
            if str(self.revedaPluginPathObj) not in sys.path:
                sys.path.append(str(self.revedaPluginPathObj))

            # Persist to .env file
            self.update_env_file("REVEDA_PLUGIN_PATH", str(self.revedaPluginPathObj))
            self.logger.info(f"Plugin path updated to: {self.revedaPluginPathObj}")

    def updateVaModulesPath(self, newPath: str):
        """Update plugin path and persist to .env file"""
        if newPath:
            # Update environment variable
            os.environ["REVEDA_VA_MODULE_PATH"] = str(Path(newPath).resolve())

            # Persist to .env file
            self.update_env_file(
                "REVEDA_VA_MODULE_PATH", os.environ["REVEDA_VA_MODULE_PATH"]
            )
            self.logger.info(
                f"Central Verilog-A module repository path: {os.environ['REVEDA_VA_MODULE_PATH']}"
            )

    def update_env_file(self, key, value):
        """Update or add environment variable in .env file"""
        env_file = self.basePath / ".env"
        lines = []

        # Read existing .env file if it exists
        if env_file.exists():
            with env_file.open("r") as f:
                lines = f.readlines()

        # Update or add the key-value pair
        keyFound = False
        for i, line in enumerate(lines):
            if line.strip().startswith(f"{key}="):
                lines[i] = f"{key}={value}\n"
                keyFound = True
                break

        if not keyFound:
            lines.append(f"{key}={value}\n")

        # Write back to .env file
        with env_file.open("w") as f:
            f.writelines(lines)


def main():
    app = revedaApp(sys.argv)
    style_map = {"Windows": "Fusion", "Linux": "Fusion", "Darwin": "macOS"}
    style = style_map.get(platform.system())
    if style:
        app.setStyle(style)
        print(f"Applied {style} style")
    app.appMainW.setWindowTitle("Revolution EDA")
    console = app.appMainW.centralW.console
    with redirect_stdout(console), redirect_stderr(pcon.Redirect(console.errorwrite)):
        app.appMainW.show()
        return app.exec()


if __name__ == "__main__":
    main()



import json
import pathlib
import shutil

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (QMainWindow, QDialog, QApplication, )

import revedaEditor.gui.fileDialogues as fd
from revedaEditor.backend import libraryModelView as lmview, hdlBackEnd as hdl, \
    dataDefinitions as ddef, libraryMethods as libm, libBackEnd as scb
from revedaEditor.fileio.createSymbols import createVaSymbol


def importVerilogaModule(viewT: ddef.viewNameTuple, filePath: str):
    appMainW = QApplication.instance().appMainW
    libraryView = appMainW.libraryBrowser.designView
    libraryModel = libraryView.libraryModel
    # Open the import dialog
    importDlg = fd.importVerilogaCellDialogue(libraryModel, appMainW.libraryBrowser)
    importDlg.vaFileEdit.setText(filePath)
    if viewT.libraryName:
        importDlg.libNamesCB.setCurrentText(viewT.libraryName)
    if viewT.cellName:
        importDlg.cellNamesCB.setCurrentText(viewT.cellName)
    if viewT.viewName:
        importDlg.vaViewName.setText(viewT.viewName)
    else:
        # Set the default view name in the dialog
        importDlg.vaViewName.setText("veriloga")
    # Execute the import dialog and check if it was accepted
    if importDlg.exec() == QDialog.DialogCode.Accepted:
        # Create the Verilog-A object from the file path
        importedVAObj = hdl.verilogaC(pathlib.Path(importDlg.vaFileEdit.text()))

        # Create the Verilog-A view item tuple
        vaViewItemTuple = createVaView(appMainW.libraryBrowser, importDlg, libraryModel,
                                       importedVAObj)
        viewsModel = libraryView.createViewsListModel(vaViewItemTuple.cellItem)
        libraryView.viewsListView.setModel(viewsModel)
        # Check if the symbol checkbox is checked
        if importDlg.symbolCheckBox.isChecked():
            # Create the Verilog-A symbol
            createVaSymbol(appMainW.libraryBrowser, vaViewItemTuple, appMainW.libraryDict,
                           appMainW.libraryBrowser,
                           importedVAObj, )


def createVaView(parent: QMainWindow, importDlg: QDialog,
                 libraryModel: lmview.designLibrariesModel,
                 importedVaObj: hdl.verilogaC, ) -> ddef.viewItemTuple:
    """
    Create a new Verilog-A view.

    Args:
        parent (QMainWindow): The parentW window.
        importDlg (QDialog): The import dialog window.
        libraryModel (edw.designLibrariesModel): The model for the design libraries.
        importedVaObj (hdl.verilogaC): The imported Verilog-A object.

    Returns:
        tuple: A tuple containing the library item, cell item, and Verilog-A item.
    """
    # Get the file path of the imported Verilog-A file
    importedVaFilePathObj = pathlib.Path(importDlg.vaFileEdit.text())

    # Get the selected library item
    libItem = libm.getLibItem(libraryModel, importDlg.libNamesCB.currentText())
    libItemRow = libItem.row()

    # Get the cell names in the selected library
    libCellNames = [libraryModel.item(libItemRow).child(i).cellName for i in
                    range(libraryModel.item(libItemRow).rowCount())]

    # Get the selected cell name
    cellName = importDlg.cellNamesCB.currentText().strip()

    # If the cell name is not in the library and is not empty, create a new cell
    if cellName not in libCellNames and cellName != "":
        scb.createCell(parent, libItem, cellName)

    # Get the cell item
    cellItem = libm.getCellItem(libItem, cellName)

    # Generate the new file path for the Verilog-A file
    newVaFilePathObj = cellItem.cellPath.joinpath(
        importedVaFilePathObj.name)

    # Create the Verilog-A item view
    vaItem = scb.createCellView(parent, importDlg.vaViewName.text(), cellItem)

    # Create a temporary copy of the imported Verilog-A file
    tempFilePathObj = importedVaFilePathObj.with_suffix(".temp")
    shutil.copy(importedVaFilePathObj, tempFilePathObj)

    # Copy the temporary file to the new file path
    shutil.copy(tempFilePathObj, newVaFilePathObj)

    # Remove the temporary file
    tempFilePathObj.unlink()

    # Create a list of items to be stored in the Verilog-A item data
    items = list()
    items.insert(0, {"cellView": "veriloga"})
    items.insert(1, {"filePath": str(newVaFilePathObj.name)})
    items.insert(2, {"vaModule": importedVaObj.vaModule})

    # Write the items to the Verilog-A item data file
    with vaItem.viewPath.open(mode="w") as f:
        json.dump(items, f, indent=4)

    # Return the tuple of library item, cell item, and Verilog-A item
    return ddef.viewItemTuple(libItem, cellItem, vaItem)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import importlib
import pkgutil
from pathlib import Path

from PySide6.QtGui import (QAction, QIcon)
from PySide6.QtWidgets import (QApplication, )
from revedaEditor.backend.dataDefinitions import viewItemTuple
from revedaEditor.backend.libBackEnd import viewItem

class pluginsLoader:
    def __init__(self, pluginsPath: Path):
        self.plugins = {}
        self.pluginsPathObj = pluginsPath
        self._app = QApplication.instance()
        self.pluginMenuConfig = {}

        for _, name, _ in pkgutil.iter_modules([str(self.pluginsPathObj)]):
            self._app.logger.info(f"Found plugin: {name}")
            try:
                module = importlib.import_module(name)
                self.plugins[f"{name}"] = module
            except ImportError as e:
                self._app.logger.error(f"Failed to load plugin {name}: {e}")
        self._app.logger.info(f"Loaded plugins: {list(self.plugins.keys())}")
        self._loadPluginMenus()

    def __repr__(self):
        return f'plugins({list(self.plugins.keys())})'

    def _loadPluginMenus(self):
        """Load plugin menu configurations from JSON file"""
        import json

        self.pluginMenuConfig = {}
        if not self.pluginsPathObj.exists():
            return

        for item in self.pluginsPathObj.iterdir():
            if item.is_dir():
                configPath = item / "config.json"
                if configPath.exists():
                    try:
                        with open(configPath) as f:
                            config = json.load(f)
                        self.pluginMenuConfig[item.name] = config
                    except Exception as e:
                        self._app.logger.warning(
                            f"Failed to load plugin config for {item.name}: {e}")

        self._app.logger.info(
            f"Loaded plugin menu config: {list(self.pluginMenuConfig.keys())}")

    def applyPluginMenus(self, editorWindow):
        """Apply plugin menus to an editor window"""
        if not hasattr(self, 'pluginMenuConfig'):
            return
        editorClassName = editorWindow.__class__.__name__

        for plugin_name, plugin_config in self.pluginMenuConfig.items():
            # Get plugin module
            module = self.plugins.get(plugin_name)
            if not module:
                continue

            for menuItem in plugin_config.get("menu_items", []):
                # Check applicability
                if "apply" in menuItem and editorClassName not in menuItem["apply"]:
                    continue
                # Get callback
                callback = getattr(module, menuItem["callback"], None)
                if not callback:
                    continue

                # Find target menu and add action
                if hasattr(editorWindow, menuItem['location']):
                    for action in editorWindow.menuBar().actions():
                        if action.text().replace('&', '') == menuItem["menu"]:
                            new_action = QAction(menuItem["action"], editorWindow)
                            if "text" in menuItem:
                                new_action.setText(menuItem["text"])
                            if "icon" in menuItem:
                                new_action.setIcon(QIcon(menuItem["icon"]))
                            if "shortcut" in menuItem:
                                new_action.setShortcut(menuItem["shortcut"])
                            if "checked" in menuItem:
                                new_action.setCheckable(True)
                                new_action.setChecked(menuItem["checked"])
                            new_action.triggered.connect(lambda c=False, cb=callback: cb(
                                editorWindow))
                            action.menu().addAction(new_action)
                            break

    def createCellView(self, viewItemT: viewItemTuple):

        for pluginName, pluginModule in self.plugins.items():
            if (hasattr(pluginModule, 'viewTypes') and
                    viewItemT.viewItem.viewType in getattr(pluginModule, 'viewTypes')):
                pluginModule.createCellView(viewItemT)
                return True
        else:
            self._app.logger.warning(
                f"No plugin found to open view type: "
                f"{viewItemT.viewItem.viewType}")
            return False

    def openCellView(self, viewItemT:viewItemTuple):
        for pluginName, pluginModule in self.plugins.items():
            if (hasattr(pluginModule, 'viewTypes') and
                    viewItemT.viewItem.viewType in getattr(pluginModule,
                                                       'viewTypes')):
                pluginModule.openCellView(viewItemT)
                return True
        else:
            self._app.logger.warning(
                f"No plugin found to open view type: "
                f"{viewItemT.viewItem.viewType}")
            return False



#    "Commons Clause" License Condition v1.0
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

"""
Editor factory for creating appropriate editor instances based on view type.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Type

from revedaEditor.gui.editorTypes import BaseEditor

if TYPE_CHECKING:
    import revedaEditor.backend.libBackEnd as libb
    import revedaEditor.backend.libraryModelView as lmview


class EditorFactory:
    """Factory for creating editor instances based on view type."""

    _editor_registry: Dict[str, Type[BaseEditor]] = {}

    @classmethod
    def registerEditor(cls, view_type: str,
                       editor_class: Type[BaseEditor]) -> None:
        """Register an editor class for a specific view type."""
        cls._editor_registry[view_type] = editor_class

    @classmethod
    def createEditor(cls, view_type: str, viewItem: 'libb.viewItem',
                     libraryDict: dict,
                     libraryView: 'lmview.BaseDesignLibrariesView') -> BaseEditor:
        """Create appropriate editor based on view type."""
        if view_type not in cls._editor_registry:
            raise ValueError(f"No editor registered for view type: {view_type}")

        editor_class = cls._editor_registry[view_type]
        return editor_class(viewItem, libraryDict, libraryView)

    @classmethod
    def getSupportedViewTypes(cls) -> list[str]:
        """Get list of supported view types."""
        return list(cls._editor_registry.keys())


# Registration function to be called by each editor module
def registerEditors():
    """Register all editor types. Called during module initialization."""
    # Import here to avoid circular imports
    from revedaEditor.gui.schematicEditor import schematicEditor
    from revedaEditor.gui.layoutEditor import layoutEditor
    from revedaEditor.gui.symbolEditor import symbolEditor

    EditorFactory.registerEditor("schematic", schematicEditor)
    EditorFactory.registerEditor("layout", layoutEditor)
    EditorFactory.registerEditor("symbol", symbolEditor)



#   “Commons Clause” License Condition v1.0
#  #
#   The Software is provided to you by the Licensor under the License, as defined
#   below, subject to the following condition.
#  #
#   Without limiting other conditions in the License, the grant of rights under the
#   License will not include, and the License does not grant to you, the right to
#   Sell the Software.
#  #
#   For purposes of the foregoing, “Sell” means practicing any or all of the rights
#   granted to you under the License to provide to third parties, for a fee or other
#   consideration (including without limitation fees for hosting or consulting/
#   support services related to the Software), a product or service whose value
#   derives, entirely or substantially, from the functionality of the Software. Any
#   license notice or attribution required by the License must also include this
#   Commons Clause License Condition notice.
#  #
#   Software: Revolution EDA
#   License: Mozilla Public License 2.0
#   Licensor: Revolution Semiconductor (Registered in the Netherlands)

import json
# schematic editor backend
import pathlib
import shutil
from pathlib import Path
from typing import Union, Optional

from PySide6.QtCore import (Qt, )
from PySide6.QtGui import (QStandardItem, )
from PySide6.QtWidgets import QMessageBox, QWidget


class libraryItem(QStandardItem):
    def __init__(self, libraryPath: pathlib.Path):  # path is a pathlib.Path object
        self._libraryPath = libraryPath
        self._libraryName = libraryPath.name
        super().__init__(self.libraryName)
        self.setEditable(False)
        self.setData(libraryPath, Qt.ItemDataRole.UserRole + 2)
        self.setData("library", Qt.ItemDataRole.UserRole + 1)
        self.setData(self.libraryName, Qt.ItemDataRole.UserRole + 3)

    def type(self):
        return Qt.StandardItem.UserType

    def __str__(self):
        return (
            f"library item path: {self.libraryPath}, library item name: {self.libraryName}")

    def __repr__(self):
        return f"{type(self).__name__}({self.libraryPath})"

    @property
    def libraryPath(self) -> pathlib.Path:
        return self._libraryPath

    @libraryPath.setter
    def libraryPath(self, value):
        if isinstance(value, pathlib.Path):
            self._libraryPath = value

    @property
    def libraryName(self) -> str:
        return self._libraryName


class cellItem(QStandardItem):
    def __init__(self, cellPath: pathlib.Path) -> None:
        self.cellPath = cellPath
        self._cellName = cellPath.stem
        super().__init__(self.cellName)
        self.setEditable(False)
        self.setData("cell", Qt.ItemDataRole.UserRole + 1)
        self.setData(cellPath, Qt.ItemDataRole.UserRole + 2)
        self.setData(self.cellName, Qt.ItemDataRole.UserRole + 3)

    def type(self):
        return QStandardItem.UserType + 1

    def __str__(self):
        return f"cell item path: {self.cellPath}, \ncell item name: {self.cellName}"

    def __repr__(self):
        return f"{type(self).__name__}({self.cellPath})"

    def delete(self):
        """
        delete the cell file and remove the row.
        """
        self.cellPath.unlink()
        cellRow = self.row()
        self.parent().removeRow(cellRow)
        # delete the cell's views
        for view in self.cellPath.parent.glob("*"):
            if view.is_file():
                view.unlink()

    @property
    def cellName(self):
        return self._cellName

    def clone(self):
        """
        Clone the cell item and return a new cell item with the same path.
        """
        newCellItem = cellItem(self.cellPath)
        newCellItem.setData('clone', Qt.ItemDataRole.UserRole + 4)
        newCellItem.setData(self, Qt.ItemDataRole.UserRole + 10)
        return newCellItem

    def refreshCellViews(self) -> None:
        """Refresh all view children of a cell item from filesystem."""
        self.removeRows(0, self.rowCount())
        cellPath = self.cellPath
        for viewPath in cellPath.glob("*.json"):
            self.appendRow(viewItem(viewPath))


class viewItem(QStandardItem):
    def __init__(self, viewPath: pathlib.Path) -> None:
        self.viewPath = viewPath
        self.viewName = viewPath.stem
        super().__init__(self.viewPath.stem)
        self.setEditable(False)
        self.setData("view", Qt.ItemDataRole.UserRole + 1)
        # set the data to the item to be the path to the view.
        self.setData(viewPath, Qt.ItemDataRole.UserRole + 2)
        self.setData(self.viewName, Qt.ItemDataRole.UserRole + 3)

    def type(self):
        return QStandardItem.UserType + 1

    def __str__(self):
        return f"view item path: {self.viewPath}, view item name: {self.viewName}"

    def __repr__(self):
        return f"{type(self).__name__}(pathlib.Path({self.viewPath}))"

    def delete(self):
        """
        delete the view file and remove the row.
        """
        self.viewPath.unlink()
        viewRow = self.row()
        self.parent().removeRow(viewRow)

    @property
    def viewType(self):
        if "schematic" in self.viewPath.stem:
            return "schematic"
        elif "symbol" in self.viewPath.stem:
            return "symbol"
        elif "veriloga" in self.viewPath.stem:
            return "veriloga"
        elif "config" in self.viewPath.stem:
            return "config"
        elif "xyce" in self.viewPath.stem:
            return "xyce"
        elif "spice" in self.viewPath.stem:
            return "spice"
        elif "myhdl" in self.viewPath.stem:
            return "myhdl"
        elif "layout" in self.viewPath.stem:
            return "layout"
        elif "pcell" in self.viewPath.stem:
            return "pcell"
        elif "revbench" in self.viewPath.stem:
            return "revbench"
        else:
            return None

    def clone(self):
        """
        Clone the view item and return a new view item with the same path.
        """
        newViewItem = viewItem(self.viewPath)
        newViewItem.setEditable(False)
        newViewItem.setData('clone', Qt.ItemDataRole.UserRole + 4)
        newViewItem.setData(self, Qt.ItemDataRole.UserRole + 10)
        return newViewItem

def createLibrary(parent, model, libraryDir: str, libraryName: str) -> Union[
    libraryItem, None]:
    """
    Create a library item with the given parameters and add it to the model.
    If the library name is empty, show a warning message.
    If the library already exists, show a warning message.
    Log the creation of the library item.
    Return the newly created library item.
    """
    if not libraryName.strip():
        QMessageBox.warning(parent, "Error", "Please enter a library name")
        return None

    libraryPath = Path(libraryDir).joinpath(libraryName)
    if libraryPath.exists():
        QMessageBox.warning(parent, "Error", "Library already exits.")
        return None

    newLibraryItem = createNewLibraryItem(model, libraryPath)
    parent.logger.info(f"Created {libraryPath}")
    return newLibraryItem


def createNewLibraryItem(model, libraryPath):
    libraryPath.mkdir()
    newLibraryItem = libraryItem(libraryPath)
    newLibraryItem.setData(libraryPath, Qt.ItemDataRole.UserRole + 2)
    newLibraryItem.setData("library", Qt.ItemDataRole.UserRole + 1)
    model.appendRow(newLibraryItem)
    return newLibraryItem


def createCell(parent, selectedLib: libraryItem, cellName: str) -> Union[cellItem, None]:
    if selectedLib.data(Qt.ItemDataRole.UserRole + 1) == "library":
        selectedLibPath = selectedLib.data(Qt.ItemDataRole.UserRole + 2)
        cellPath = selectedLibPath.joinpath(cellName)
        if cellName.strip() == "":
            QMessageBox.warning(parent, "Error", "Please enter a cell name")
            return None
        elif cellPath.exists():
            QMessageBox.warning(parent, "Error", "Cell already exits. Delete cell first.")
            return None
        else:
            newCellItem = createNewCellItem(selectedLib, cellPath)
            parent.logger.info(f"Created {cellName} cell at {str(cellPath)}")
            return newCellItem


def createNewCellItem(selectedLib, cellPath):
    cellPath.mkdir(mode=0o755, parents=True, exist_ok=True)
    newCellItem = cellItem(cellPath)
    selectedLib.appendRow(newCellItem)
    return newCellItem


def createCellView(parent: QWidget, viewName: str, cellItem: cellItem) -> viewItem:
    """
    Create a cell view with the given view name and cell item.
    If the view name is empty, show a warning message.
    If the view path exists, replace the cell view and show a warning message.
    Create an empty cell view path and append the new view item to the cell item.
    Return the new view item.
    """
    if viewName.strip() == "":
        QMessageBox.warning(parent, "Error", "Please enter a view name")
        return None
    viewPath = cellItem.cellPath.joinpath(f"{viewName}.json")
    if viewPath.exists():
        parent.logger.warning("Replacing the cell view.")
        for row in range(cellItem.rowCount()):
            child = cellItem.child(row)
            if isinstance(child, viewItem):
                if child.viewName == viewName:
                    cellItem.removeRow(row)
                    break
            elif child and child.text() == viewName:
                # Handle case where child is a QStandardItem
                cellItem.removeRow(row)
                break
    newViewItem = createCellviewItem(viewName, viewPath)
    parent.logger.warning(f"Created {viewName} at {str(viewPath)}")
    cellItem.appendRow(newViewItem)
    return newViewItem


def createCellviewItem(viewName, viewPath)->viewItem:
    newViewItem = viewItem(viewPath)
    viewPath.touch()  # create empty cell view path
    items = list()
    if "schematic" in viewName:
        items.insert(0, {"viewType": "schematic"})
    elif "symbol" in viewName:
        items.insert(0, {"viewType": "symbol"})
    elif "layout" in viewName:
        items.insert(0, {"viewType": "layout"})
    elif "pcell" in viewName:
        items.insert(0, {"viewType": "pcell"})
    elif "spice" in viewName:
        items.insert(0, {"viewType": "spice"})
    elif "veriloga" in viewName:
        items.insert(0, {"viewType": "veriloga"})
    elif "config" in viewName:
        items.insert(0, {"viewType": "config"})
    # elif "revbench" in viewName:
    #     items.insert(0, {"viewType": "revbench"})
    with viewPath.open(mode="w") as f:
        json.dump(items, f, indent=4)
    return newViewItem


def copyCell(parent, model, origCellItem: cellItem, copyName, selectedLibPath) -> tuple[
    bool, Optional[cellItem]]:
    """Copy a cell to a new location with a new name."""
    cellPath = origCellItem.data(Qt.ItemDataRole.UserRole + 2)
    copyName = copyName or "newCell"
    copyPath = selectedLibPath.joinpath(copyName)

    if copyPath.exists():
        QMessageBox.warning(parent, "Error", "Cell already exits.")
        return False, None

    try:
        # Copy the cell directory
        shutil.copytree(cellPath, copyPath)

        # Create new cell and add to library
        libraryItem = model.findItems(selectedLibPath.name, flags=Qt.MatchExactly)[0]
        newCellItem = cellItem(copyPath)

        # Add JSON files as view items
        viewItems = [viewItem(p) for p in copyPath.iterdir() if p.suffix == ".json"]
        if viewItems:
            newCellItem.appendRows(viewItems)

        libraryItem.appendRow(newCellItem)
        return True, newCellItem

    except (FileNotFoundError, IndexError) as e:
        QMessageBox.warning(parent, "Error", f"Failed to copy cell: {str(e)}")
        return False, None

def renameCell(parent, oldCell: cellItem, newName: str) -> bool:
    """
    Function to rename a cell in the parentW with a new name.
    Parameters:
    - parentW: the parentW of the cell
    - oldCell: the cell to be renamed
    - newName: the new name for the cell
    Returns:
    - bool: True if the cell is successfully renamed, False otherwise
    """
    cellPath = oldCell.data(Qt.ItemDataRole.UserRole + 2)
    if newName.strip() == "":
        QMessageBox.warning(parent, "Error", "Please enter a cell name")
        return False
    else:
        cellPath.rename(cellPath.parent / newName)
        oldCell.setText(newName)
        return True






#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import json
from typing import Dict, Any

from PySide6.QtCore import QPointF

import revedaEditor.common.net as net
import revedaEditor.common.shapes as shp


class schematicEncoder(json.JSONEncoder):
    def default(self, item: Any) -> Dict[str, Any]:
        if isinstance(item, shp.schematicSymbol):
            return self._encodeSchematicSymbol(item)
        elif isinstance(item, net.schematicNet):
            return self._encodeSchematicNet(item)
        elif isinstance(item, shp.schematicPin):
            return self._encodeSchematicPin(item)
        elif isinstance(item, shp.text):
            return self._encodeText(item)
        return super().default(item)

    def _encodeSchematicSymbol(self, item: shp.schematicSymbol) -> Dict[str, Any]:
        item_label_dict = (
            item.labelDict if item.draft
            else {label.labelName: [label.labelValue, label.labelVisible]
                  for label in item.labels.values()}
        )
        scene_origin = item.scene().origin
        return {
            "type": "sys",
            "lib": item.libraryName,
            "cell": item.cellName,
            "view": item.viewName,
            "nam": item.instanceName,
            "ic": item.counter,
            "ld": item_label_dict,
            "loc": self._subtract_point(item.scenePos(), scene_origin),
            "ang": item.angle,
            "ign": int(item.netlistIgnore),
            "br": item.boundingRect().getCoords(),
            "fl": item.flipTuple,
        }

    def _encodeSchematicNet(self, item: net.schematicNet) -> Dict[str, Any]:
        scene_origin = item.scene().origin
        return {
            "type": "scn",
            "st": self._subtract_point(item.mapToScene(item.draftLine.p1()), scene_origin),
            "end": self._subtract_point(item.mapToScene(item.draftLine.p2()), scene_origin),
            "nam": item.name,
            "ns": item.nameStrength.value,
        }

    def _encodeSchematicPin(self, item: shp.schematicPin) -> Dict[str, Any]:
        return {
            "type": "scp",
            "st": self._subtract_point(item.mapToScene(item.start), item.scene().origin),
            "pn": item.pinName,
            "pd": item.pinDir,
            "pt": item.pinType,
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    def _encodeText(self, item: shp.text) -> Dict[str, Any]:
        return {
            "type": "txt",
            "st": self._subtract_point(item.mapToScene(item.start), item.scene().origin),
            "tc": item.textContent,
            "ff": item.fontFamily,
            "fs": item.fontStyle,
            "th": item.textHeight,
            "ta": item.textAlignment,
            "to": item.textOrient,
            "ang": item.angle,
            "fl": item.flipTuple,
        }

    @staticmethod
    def _subtract_point(point: QPointF, origin: QPointF) -> tuple:
        return (point - origin).toTuple()



#     "Commons Clause" License Condition v1.0
#    #
#     The Software is provided to you by the Licensor under the License, as defined
#     below, subject to the following condition.
#  #
#     Without limiting other conditions in the License, the grant of rights under the
#     License will not include, and the License does not grant to you, the right to
#     Sell the Software.
#  #
#     For purposes of the foregoing, "Sell" means practicing any or all of the rights
#     granted to you under the License to provide to third parties, for a fee or other
#     consideration (including without limitation fees for hosting) a product or service whose value
#     derives, entirely or substantially, from the functionality of the Software. Any
#     license notice or attribution required by the License must also include this
#     Commons Clause License Condition notice.
#  #
#    Add-ons and extensions developed for this software may be distributed
#    under their own separate licenses.
#  #
#     Software: Revolution EDA
#     License: Mozilla Public License 2.0
#     Licensor: Revolution Semiconductor (Registered in the Netherlands)


from typing import TYPE_CHECKING, Union

from PySide6.QtCore import (
    QLineF,
)
from PySide6.QtWidgets import (
    QApplication,
    QButtonGroup,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QGraphicsItem,
    QGroupBox,
    QHBoxLayout,
    QRadioButton,
    QVBoxLayout,
)
from quantiphy.quantiphy import Quantity

import revedaEditor.gui.editFunctions as edf
from revedaEditor.backend.pdkLoader import importPDKModule
from revedaEditor.common.shapes import alignLine

process = importPDKModule("process")
if TYPE_CHECKING:
    from revedaEditor.gui.editorWindow import editorWindow


# Assuming layout units are given, this is the conversion factor between


class alignItemsDialogue(QDialog):
    def __init__(self, parentW: "editorWindow"):
        super().__init__(parentW)
        self.editor = parentW
        self.scene = parentW.centralW.scene
        sceneName = ''
        sceneType = self.scene.__class__.__name__
        match sceneType:
            case "schematicScene":
                sceneName = "Schematic"
            case "symbolScene":
                sceneName = "Symbol"
            case "layoutScene":
                sceneName = "Layout"
        self.alignLine = QLineF()
        self.setWindowTitle(f"Align {sceneName} items")
        self.setMinimumWidth(340)
        self.setMinimumHeight(400)
        mainLayout = QVBoxLayout()
        alignMethodGroup = QGroupBox("Align Method")
        alignMethodLayout = QHBoxLayout(alignMethodGroup)
        alignMethodGroup.setLayout(alignMethodLayout)
        alignButtonGroup = QButtonGroup(alignMethodGroup)
        self.alignEdgesButton = QRadioButton("Align Edges")
        self.alignEdgesButton.setChecked(True)
        self.alignEdgesButton.pressed.connect(self.edgeAlignmentMessageShow)
        alignButtonGroup.addButton(self.alignEdgesButton)
        self.alignLineButton = QRadioButton("Align To Line")
        self.alignLineButton.pressed.connect(self.lineAlignmentMessageShow)
        alignButtonGroup.addButton(self.alignLineButton)
        alignMethodLayout.addWidget(self.alignEdgesButton)
        alignMethodLayout.addWidget(self.alignLineButton)
        mainLayout.addWidget(alignMethodGroup)
        alignDirectionGroup = QGroupBox("Align Direction")
        alignButtonGroup = QButtonGroup(alignDirectionGroup)
        alignLayout = QHBoxLayout(alignDirectionGroup)
        self.horizontalAlignButton = QRadioButton("Horizontal")
        self.horizontalAlignButton.pressed.connect(self.showHorizonalAlignments)
        alignButtonGroup.addButton(self.horizontalAlignButton)
        self.verticalAlignButton = QRadioButton("Vertical")
        self.verticalAlignButton.pressed.connect(self.showVerticalAlignments)
        alignButtonGroup.addButton(self.verticalAlignButton)
        self.horizontalAlignButton.setChecked(True)
        alignLayout.addWidget(self.horizontalAlignButton)
        alignLayout.addWidget(self.verticalAlignButton)
        alignDirectionGroup.setLayout(alignLayout)
        mainLayout.addWidget(alignDirectionGroup)
        alignOptionsGroup = QGroupBox("Align Options")
        alignOptionsLayout = QFormLayout(alignOptionsGroup)
        alignOptionsGroup.setLayout(alignOptionsLayout)
        self.alignSpacingTextEdit = edf.shortLineEdit()
        alignOptionsLayout.addRow(edf.boldLabel("Spacing:"), self.alignSpacingTextEdit)
        mainLayout.addWidget(alignOptionsGroup)

        self.horizontalAlignGroup = QGroupBox("Horizontal Align Options")
        horizontalAlignLayout = QHBoxLayout(self.horizontalAlignGroup)
        self.horizontalAlignGroup.setLayout(horizontalAlignLayout)
        horizontalButtonGroup = QButtonGroup(self.horizontalAlignGroup)
        self.topHorizontalAlignButton = QRadioButton("Top")
        self.topHorizontalAlignButton.setChecked(True)
        horizontalButtonGroup.addButton(self.topHorizontalAlignButton)
        self.centreHorizontalAlignButton = QRadioButton("Centre")
        horizontalButtonGroup.addButton(self.centreHorizontalAlignButton)
        self.bottomHorizontalAlignButton = QRadioButton("Bottom")
        horizontalButtonGroup.addButton(self.bottomHorizontalAlignButton)
        horizontalAlignLayout.addWidget(self.topHorizontalAlignButton)
        horizontalAlignLayout.addWidget(self.centreHorizontalAlignButton)
        horizontalAlignLayout.addWidget(self.bottomHorizontalAlignButton)
        mainLayout.addWidget(self.horizontalAlignGroup)

        self.verticalAlignGroup = QGroupBox("Vertical Align Options")
        verticalAlignLayout = QHBoxLayout(self.verticalAlignGroup)
        self.verticalAlignGroup.setLayout(verticalAlignLayout)
        verticalButtonGroup = QButtonGroup(self.verticalAlignGroup)
        self.leftVerticalAlignButton = QRadioButton("Left")
        self.leftVerticalAlignButton.setChecked(True)
        verticalButtonGroup.addButton(self.leftVerticalAlignButton)
        self.centerVerticalAlignButton = QRadioButton("Centre")
        verticalButtonGroup.addButton(self.centerVerticalAlignButton)
        self.rightVerticalAlignButton = QRadioButton("Right")
        verticalButtonGroup.addButton(self.rightVerticalAlignButton)
        verticalAlignLayout.addWidget(self.leftVerticalAlignButton)
        verticalAlignLayout.addWidget(self.centerVerticalAlignButton)
        verticalAlignLayout.addWidget(self.rightVerticalAlignButton)
        self.verticalAlignGroup.hide()
        mainLayout.addWidget(self.verticalAlignGroup)

        QBtn = (
                QDialogButtonBox.StandardButton.Apply
                | QDialogButtonBox.StandardButton.Ok
                | QDialogButtonBox.StandardButton.Cancel
        )
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.rejected.connect(self.reject)

        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)
        self.editor.messageLine.setText("Select items to be aligned.")
        self.show()

    def edgeAlignmentMessageShow(self):
        self.editor.messageLine.setText("Select items to be aligned.")

    def lineAlignmentMessageShow(self):
        self.editor.messageLine.setText(
            "Select Reference Line and items to be aligned."
        )

    def showHorizonalAlignments(self):
        self.horizontalAlignGroup.show()
        self.verticalAlignGroup.hide()

    def showVerticalAlignments(self):
        self.horizontalAlignGroup.hide()
        self.verticalAlignGroup.show()


def alignItemsEdgeHorizontally(
        selectedItems: list[QGraphicsItem], alignment: str, spacing: Union[float, str]
):
    # if len(selectedItems) < 2:
    #     return
    # Sort items by x position for distribution
    sorted_items = sorted(
        selectedItems, key=lambda item: item.sceneBoundingRect().left()
    )
    match alignment:
        case "top":
            target_y = min(item.sceneBoundingRect().top() for item in selectedItems)
            for item in sorted_items:
                current_top = item.sceneBoundingRect().top()
                item.moveBy(0, target_y - current_top)
        case "center":
            target_y = sum(
                item.sceneBoundingRect().center().y() for item in selectedItems
            ) / len(selectedItems)
            for item in sorted_items:
                current_center = item.sceneBoundingRect().center().y()
                item.moveBy(0, target_y - current_center)
        case "bottom":
            target_y = max(item.sceneBoundingRect().bottom() for item in selectedItems)
            for item in sorted_items:
                current_bottom = item.sceneBoundingRect().bottom()
                item.moveBy(0, target_y - current_bottom)
    # If no spacing value is given, return
    if spacing == "":
        return
    # Distribute with spacing
    current_x = sorted_items[0].sceneBoundingRect().left()
    for item in sorted_items[1:]:
        current_x += sorted_items[
                         sorted_items.index(item) - 1
                         ].sceneBoundingRect().width() + float(spacing)
        item.moveBy(current_x - item.sceneBoundingRect().left(), 0)


def alignItemsEdgeVertically(
        selectedItems: list[QGraphicsItem], alignment: str, spacing: Union[float, str]
):
    # if len(selectedItems) < 2:
    #     return
    # Sort items by x position for distribution
    sorted_items = sorted(
        selectedItems, key=lambda item: item.sceneBoundingRect().left()
    )
    match alignment:
        case "left":
            target_x = min(item.sceneBoundingRect().left() for item in selectedItems)
            for item in sorted_items:
                current_left = item.sceneBoundingRect().left()
                item.moveBy(target_x - current_left, 0)
        case "center":
            target_x = sum(
                item.sceneBoundingRect().center().x() for item in selectedItems
            ) / len(selectedItems)
            for item in sorted_items:
                current_center = item.sceneBoundingRect().center().x()
                item.moveBy(target_x - current_center, 0)
        case "right":
            target_x = max(item.sceneBoundingRect().right() for item in selectedItems)
            for item in sorted_items:
                current_right = item.sceneBoundingRect().bottom()
                item.moveBy(target_x - current_right, 0)

    if spacing == "":
        return
    # Distribute with spacing
    current_y = sorted_items[0].sceneBoundingRect().top()
    for item in sorted_items[1:]:
        current_y += sorted_items[
                         sorted_items.index(item) - 1
                         ].sceneBoundingRect().height() + float(spacing)
        item.moveBy(0, current_y - item.sceneBoundingRect().top())


def alignItemsLineHorizontally(
        selectedItems: list[QGraphicsItem],
        alignment: str,
        spacing: Union[float, str],
        target_y: int,
):
    # Sort items by x position for distribution
    sorted_items = sorted(
        selectedItems, key=lambda item: item.sceneBoundingRect().left()
    )
    match alignment:
        case "top":
            for item in sorted_items:
                current_top = item.sceneBoundingRect().top()
                item.moveBy(0, target_y - current_top)
        case "center":
            for item in sorted_items:
                current_center = item.sceneBoundingRect().center().y()
                item.moveBy(0, target_y - current_center)
        case "bottom":
            for item in sorted_items:
                current_bottom = item.sceneBoundingRect().bottom()
                item.moveBy(0, target_y - current_bottom)
    # If no spacing value is given, return
    if spacing == "":
        return
    # Distribute with spacing
    current_x = sorted_items[0].sceneBoundingRect().left()
    for item in sorted_items[1:]:
        current_x += sorted_items[
                         sorted_items.index(item) - 1
                         ].sceneBoundingRect().width() + float(spacing)
        item.moveBy(current_x - item.sceneBoundingRect().left(), 0)


def alignItemsLineVertically(
        selectedItems: list[QGraphicsItem],
        alignment: str,
        spacing: Union[float, str],
        target_x: int,
):
    # if len(selectedItems) < 2:
    #     return
    # Sort items by x position for distribution
    sorted_items = sorted(
        selectedItems, key=lambda item: item.sceneBoundingRect().left()
    )
    match alignment:
        case "left":
            for item in sorted_items:
                current_left = item.sceneBoundingRect().left()
                item.moveBy(target_x - current_left, 0)
        case "center":
            for item in sorted_items:
                current_center = item.sceneBoundingRect().center().x()
                item.moveBy(target_x - current_center, 0)
        case "right":
            for item in sorted_items:
                current_right = item.sceneBoundingRect().bottom()
                item.moveBy(target_x - current_right, 0)

    if spacing == "":
        return
    # Distribute with spacing
    current_y = sorted_items[0].sceneBoundingRect().top()
    for item in sorted_items[1:]:
        current_y += sorted_items[
                         sorted_items.index(item) - 1
                         ].sceneBoundingRect().height() + float(spacing)
        item.moveBy(0, current_y - item.sceneBoundingRect().top())


def handleAlignAction(dlg: alignItemsDialogue, closeDialog: bool):
    # handles apply or OK key presses in the dialogue

    selectedItems = dlg.scene.selectedItems()
    spacingText = dlg.alignSpacingTextEdit.text().strip()
    if dlg.editor.__class__.__name__ in ("schematicEditor", "symbolEditor"):
        spacing = int(Quantity(dlg.alignSpacingTextEdit.text()).real)
    elif dlg.editor.__class__.__name__ in ("layoutEditor"):
        spacing = (int(Quantity(spacingText).real * process.dbu) if spacingText != ""
                   else "")

    if dlg.alignEdgesButton.isChecked():
        if len(selectedItems) < 2:
            dlg.scene.logger.error("Less than 2 items are selected.")
            return
        if dlg.horizontalAlignButton.isChecked():
            if dlg.topHorizontalAlignButton.isChecked():
                alignItemsEdgeHorizontally(selectedItems, "top", spacing)
            elif dlg.centreHorizontalAlignButton.isChecked():
                alignItemsEdgeHorizontally(selectedItems, "center", spacing)
            elif dlg.bottomHorizontalAlignButton.isChecked():
                alignItemsEdgeHorizontally(selectedItems, "bottom", spacing)
        else:
            if dlg.leftVerticalAlignButton.isChecked():
                alignItemsEdgeVertically(selectedItems, "left", spacing)
            elif dlg.centerVerticalAlignButton.isChecked():
                alignItemsEdgeVertically(selectedItems, "center", spacing)
            elif dlg.rightVerticalAlignButton.isChecked():
                alignItemsEdgeVertically(selectedItems, "right", spacing)
    elif dlg.alignLineButton.isChecked() and not dlg.alignLine.isNull():
        if dlg.horizontalAlignButton.isChecked():
            target_y = int(dlg.alignLine.y1())
            if dlg.topHorizontalAlignButton.isChecked():
                alignItemsLineHorizontally(selectedItems, "top", spacing, target_y)
            elif dlg.centreHorizontalAlignButton.isChecked():
                alignItemsLineHorizontally(selectedItems, "center", spacing, target_y)
            elif dlg.bottomHorizontalAlignButton.isChecked():
                alignItemsLineHorizontally(selectedItems, "bottom", spacing, target_y)
        else:
            target_x = int(dlg.alignLine.x1())
            if dlg.leftVerticalAlignButton.isChecked():
                alignItemsLineVertically(selectedItems, "left", spacing, target_x)
            elif dlg.centerVerticalAlignButton.isChecked():
                alignItemsLineVertically(selectedItems, "center", spacing, target_x)
            elif dlg.rightVerticalAlignButton.isChecked():
                alignItemsLineVertically(selectedItems, "right", spacing, target_x)
        dlg.scene.newAlignLine = None
    if closeDialog:
        dlg.scene.removeItem(dlg.scene.newAlignLine)
        dlg.scene.undoStack.removeLastCommand()
        dlg.accept()


def startLineAlign(scene, dlg: alignItemsDialogue):
    scene.editorWindow.messageLine.setText("Start Drawing Alignment Reference Line.")
    scene.editModes.setMode("alignItems")


def startEdgeAlign(scene, dlg: alignItemsDialogue):
    scene.editorWindow.messageLine.setText("Select Items will be aligned.")
    scene.clearSelection()
    scene.deselectAll()
    scene.editModes.setMode("selectItem")


def alignToLine(newAlignLine: alignLine):
    scene = newAlignLine.scene()
    openAlignDialogues = [
        w
        for w in QApplication.topLevelWidgets()
        if isinstance(w, alignItemsDialogue) and w.isVisible() and w.scene == scene
    ]
    # there must be only one dialogue that meets the criteria.
    if openAlignDialogues:
        openAlignDialogues[0].alignLine = newAlignLine.draftLine
        scene.editModes.setMode("selectItems")
